# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# It's based on oslo.i18n usage in OpenStack Keystone project and
# recommendations from
# https://docs.openstack.org/oslo.i18n/latest/user/usage.html

"""Utilities and helper functions."""
import base64
import binascii
import contextlib
import eventlet
import functools
import inspect
import math
import mimetypes

from oslo_concurrency import lockutils
from oslo_concurrency import processutils
from oslo_context import context as common_context
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import strutils
import pecan
import six

from zun.api import utils as api_utils
from zun.common import clients
from zun.common import consts
from zun.common.docker_image import reference as docker_image
from zun.common import exception
from zun.common.i18n import _
from zun.common import privileged
import zun.conf
from zun.network import neutron
from zun import objects

CONF = zun.conf.CONF
LOG = logging.getLogger(__name__)
NETWORK_ATTACH_EXTERNAL = 'network:attach_external_network'

synchronized = lockutils.synchronized_with_prefix(consts.NAME_PREFIX)

VALID_STATES = {
    'commit': [consts.RUNNING, consts.STOPPED, consts.PAUSED],
    'delete': [consts.CREATED, consts.ERROR, consts.STOPPED, consts.DELETED,
               consts.DEAD],
    'delete_force': [consts.CREATED, consts.CREATING, consts.ERROR,
                     consts.RUNNING, consts.STOPPED, consts.UNKNOWN,
                     consts.DELETED, consts.DEAD, consts.RESTARTING,
                     consts.REBUILDING, consts.DELETING],
    'delete_after_stop': [consts.RUNNING, consts.CREATED, consts.ERROR,
                          consts.STOPPED, consts.DELETED, consts.DEAD],
    'start': [consts.CREATED, consts.STOPPED, consts.ERROR],
    'stop': [consts.RUNNING],
    'reboot': [consts.CREATED, consts.RUNNING, consts.STOPPED, consts.ERROR],
    'rebuild': [consts.CREATED, consts.RUNNING, consts.STOPPED, consts.ERROR],
    'pause': [consts.RUNNING],
    'unpause': [consts.PAUSED],
    'kill': [consts.RUNNING],
    'execute': [consts.RUNNING],
    'execute_resize': [consts.RUNNING],
    'update': [consts.CREATED, consts.RUNNING, consts.STOPPED, consts.PAUSED],
    'attach': [consts.RUNNING],
    'resize': [consts.RUNNING],
    'top': [consts.RUNNING],
    'get_archive': [consts.CREATED, consts.PAUSED, consts.RUNNING,
                    consts.STOPPED],
    'put_archive': [consts.CREATED, consts.PAUSED, consts.RUNNING,
                    consts.STOPPED],
    'logs': [consts.CREATED, consts.ERROR, consts.PAUSED, consts.RUNNING,
             consts.STOPPED, consts.UNKNOWN],
    'stats': [consts.RUNNING],
    'add_security_group': [consts.CREATED, consts.RUNNING, consts.STOPPED,
                           consts.PAUSED],
    'remove_security_group': [consts.CREATED, consts.RUNNING, consts.STOPPED,
                              consts.PAUSED],
    'resize_container': [consts.CREATED, consts.RUNNING, consts.STOPPED,
                         consts.PAUSED]
}

VALID_CONTAINER_FILED = {
    'name': 'name',
    'image': 'image',
    'command': 'command',
    'args': 'args',
    'resources': 'resources',
    'ports': 'ports',
    'volumeMounts': 'volumeMounts',
    'env': 'environment',
    'workDir': 'workdir',
    'imagePullPolicy': 'image_pull_policy',
    'tty': 'tty',
    'stdin': 'interactive',
}

VALID_CAPSULE_FIELD = {
    'restartPolicy': 'restart_policy',
}

VALID_CAPSULE_RESTART_POLICY = {
    'Never': 'no',
    'Always': 'always',
    'OnFailure': 'on-failure',
}


def validate_container_state(container, action):
    if container.status not in VALID_STATES[action]:
        raise exception.InvalidStateException(
            id=container.uuid,
            action=action,
            actual_state=container.status)


def validate_image_driver(image_driver):
    if image_driver not in CONF.image_driver_list:
        detail = _("Invalid input for image_driver, "
                   "it should be within the image drivers list")
        raise exception.ValidationError(detail=detail)


def safe_rstrip(value, chars=None):
    """Removes trailing characters from a string if that does not make it empty

    :param value: A string value that will be stripped.
    :param chars: Characters to remove.
    :return: Stripped value.

    """
    if not isinstance(value, six.string_types):
        LOG.warning(
            "Failed to remove trailing character. Returning original object. "
            "Supplied object is not a string: %s.", value)
        return value

    return value.rstrip(chars) or value


def _do_allow_certain_content_types(func, content_types_list):
    # Allows you to bypass pecan's content-type restrictions
    cfg = pecan.util._cfg(func)
    cfg.setdefault('content_types', {})
    cfg['content_types'].update((value, '')
                                for value in content_types_list)
    return func


def allow_certain_content_types(*content_types_list):
    def _wrapper(func):
        return _do_allow_certain_content_types(func, content_types_list)
    return _wrapper


def allow_all_content_types(f):
    return _do_allow_certain_content_types(f, mimetypes.types_map.values())


def parse_image_name(image, driver=None, registry=None):
    image_parts = docker_image.Reference.parse(image)

    image_repo = image_parts['name']
    if driver is None:
        driver = CONF.default_image_driver
    if driver == 'glance':
        image_tag = ''
        return image_repo, image_tag

    image_tag = 'latest'
    if image_parts['tag']:
        image_tag = image_parts['tag']

    domain, _ = image_parts.split_hostname()
    if not domain:
        if registry:
            image_repo = '%s/%s' % (registry.domain, image_repo)
        elif CONF.docker.default_registry:
            image_repo = '%s/%s' % (CONF.docker.default_registry, image_repo)

    return image_repo, image_tag


def spawn_n(func, *args, **kwargs):
    """Passthrough method for eventlet.spawn_n.

    This utility exists so that it can be stubbed for testing without
    interfering with the service spawns.

    It will also grab the context from the threadlocal store and add it to
    the store on the new thread.  This allows for continuity in logging the
    context when using this method to spawn a new thread.
    """
    _context = common_context.get_current()

    @functools.wraps(func)
    def context_wrapper(*args, **kwargs):
        # NOTE: If update_store is not called after spawn_n it won't be
        # available for the logger to pull from threadlocal storage.
        if _context is not None:
            _context.update_store()
        func(*args, **kwargs)

    eventlet.spawn_n(context_wrapper, *args, **kwargs)


def translate_exception(function):
    """Wraps a method to catch exceptions.

    If the exception is not an instance of ZunException,
    translate it into one.
    """

    @functools.wraps(function)
    def decorated_function(self, context, *args, **kwargs):
        try:
            return function(self, context, *args, **kwargs)
        except Exception as e:
            if not isinstance(e, exception.ZunException):
                LOG.exception("Unexpected error: %s", six.text_type(e))
                e = exception.ZunException("Unexpected error: %s"
                                           % six.text_type(e))
                raise e
            raise

    return decorated_function


def check_container_id(function):
    """Check container_id property of given container instance."""

    @functools.wraps(function)
    def decorated_function(*args, **kwargs):
        container = args[2]
        if getattr(container, 'container_id', None) is None:
            msg = _("Cannot operate an uncreated container.")
            raise exception.Invalid(message=msg)
        return function(*args, **kwargs)

    return decorated_function


def get_image_pull_policy(image_pull_policy, image_tag):
    if not image_pull_policy:
        image_pull_policy = 'always'
    return image_pull_policy


def should_pull_image(image_pull_policy, present):
    if image_pull_policy == 'never':
        return False
    if (image_pull_policy == 'always' or
            (image_pull_policy == 'ifnotpresent' and not present)):
        return True
    return False


def get_floating_cpu_set():
    """Parse floating_cpu_set config.

    :returns: a set of pcpu ids can be used by containers
    """

    if not CONF.floating_cpu_set:
        return None

    cpuset_ids = parse_floating_cpu(CONF.floating_cpu_set)
    if not cpuset_ids:
        raise exception.Invalid(_("No CPUs available after parsing %r") %
                                CONF.floating_cpu_set)
    return cpuset_ids


def parse_floating_cpu(spec):
    """Parse a CPU set specification.

    Each element in the list is either a single CPU number, a range of
    CPU numbers.

    :param spec: cpu set string eg "1-4,6"
    :returns: a set of CPU indexes

    """

    cpuset_ids = set()
    for rule in spec.split(','):
        range_part = rule.strip().split("-", 1)
        if len(range_part) > 1:
            try:
                start, end = [int(p.strip()) for p in range_part]
            except ValueError:
                raise exception.Invalid()
            if start < end:
                cpuset_ids |= set(range(start, end + 1))
            else:
                raise exception.Invalid()
        else:
            try:
                cpuset_ids.add(int(rule))
            except ValueError:
                raise exception.Invalid()

    return cpuset_ids


def get_security_group_ids(context, security_groups, **kwargs):
    if not security_groups:
        return None
    else:
        neutron = clients.OpenStackClients(context).neutron()
        search_opts = {'tenant_id': context.project_id}
        security_groups_list = neutron.list_security_groups(
            **search_opts).get('security_groups', [])
        security_group_ids = [item['id'] for item in security_groups_list
                              if item['name'] in security_groups
                              or item['id'] in security_groups]
        if len(security_group_ids) >= len(security_groups):
            return security_group_ids
        else:
            raise exception.ZunException(_(
                "Any of the security group in %s is not found ") %
                security_groups)


def custom_execute(*cmd, **kwargs):
    try:
        return processutils.execute(*cmd, **kwargs)
    except processutils.ProcessExecutionError as e:
        sanitized_cmd = strutils.mask_password(' '.join(cmd))
        raise exception.CommandError(cmd=sanitized_cmd,
                                     error=six.text_type(e))


def get_root_helper():
    return 'sudo zun-rootwrap %s' % CONF.rootwrap_config


@privileged.default.entrypoint
def execute_root(*cmd, **kwargs):
    # NOTE(kiennt): Set run_as_root=False because if it is set to True, the
    #               command is prefixed by the command specified in the
    #               root_helper kwargs [1]. But we use oslo.privsep instead
    #               of rootwrap so set run_as_root=False.
    # [1] https://github.com/openstack/oslo.concurrency/blob/master/oslo_concurrency/processutils.py#L218 # noqa
    return custom_execute(*cmd, shell=False, run_as_root=False, **kwargs)


def execute(*cmd, **kwargs):
    run_as_root = kwargs.pop('run_as_root', False)
    # NOTE(kiennt): Root_helper is unnecessary when use privsep,
    #               therefore pop it!
    kwargs.pop('root_helper', None)
    if run_as_root:
        return execute_root(*cmd, **kwargs)
    else:
        return custom_execute(*cmd, **kwargs)


def capsule_get_container_spec(spec_field):
    containers_spec = spec_field.get('containers')
    containers_num = len(containers_spec)
    if containers_num == 0:
        raise exception.InvalidCapsuleTemplate("Capsule need to have one "
                                               "container at least")

    for i in range(0, containers_num):
        container_spec = containers_spec[i]
        if 'image' not in container_spec.keys():
            raise exception.InvalidCapsuleTemplate("Container "
                                                   "image is needed")
        # Remap the Capsule's container fields to native Zun container fields.
        for key in list(container_spec.keys()):
            container_spec[VALID_CONTAINER_FILED[key]] = \
                container_spec.pop(key)

    init_containers_spec = spec_field.get('initContainers')
    if init_containers_spec is not None:
        for i in range(0, len(init_containers_spec)):
            container_spec = init_containers_spec[i]
            if 'image' not in container_spec.keys():
                raise exception.InvalidCapsuleTemplate("Container "
                                                       "image is needed")
            # Remap the Capsule's container fields to native Zun
            # container fields.
            for key in list(container_spec.keys()):
                container_spec[VALID_CONTAINER_FILED[key]] = \
                    container_spec.pop(key)
    else:
        init_containers_spec = []

    return containers_spec, init_containers_spec


def capsule_get_volume_spec(spec_field):
    volumes_spec = spec_field.get('volumes')
    if not volumes_spec:
        return []
    volumes_num = len(volumes_spec)

    for i in range(volumes_num):
        volume_name = volumes_spec[i].get('name')
        if volume_name is None:
            raise exception.InvalidCapsuleTemplate("Volume name "
                                                   "is needed")
        if volumes_spec[i].get('cinder'):
            cinder_spec = volumes_spec[i].get('cinder')
            volume_uuid = cinder_spec.get('volumeID')
            volume_size = cinder_spec.get('size')
            if not volume_uuid:
                if volume_size is None:
                    raise exception.InvalidCapsuleTemplate("Volume size "
                                                           "is needed")
            elif volume_uuid and volume_size:
                raise exception.InvalidCapsuleTemplate("Volume size and uuid "
                                                       "could not be set at "
                                                       "the same time")
        else:
            raise exception.InvalidCapsuleTemplate("Zun now Only support "
                                                   "Cinder volume driver")

    return volumes_spec


def is_all_projects(search_opts):
    all_projects = search_opts.get('all_projects')
    if all_projects:
        try:
            all_projects = strutils.bool_from_string(all_projects, True)
        except ValueError:
            bools = ', '.join(strutils.TRUE_STRINGS + strutils.FALSE_STRINGS)
            raise exception.InvalidValue(_('Valid all_projects values are: %s')
                                         % bools)
    else:
        all_projects = False
    return all_projects


def get_container(container_ident):
    container = api_utils.get_resource('Container', container_ident)
    if not container:
        pecan.abort(404, ('Not found; the container you requested '
                          'does not exist.'))

    return container


def get_image(image_id):
    image = api_utils.get_resource('Image', image_id)
    if not image:
        pecan.abort(404, ('Not found; the image you requested '
                          'does not exist.'))

    return image


def get_registry(registry_id):
    registry = api_utils.get_resource('Registry', registry_id)
    if not registry:
        raise exception.RegistryNotFound(registry=registry_id)

    return registry


def check_for_restart_policy(container_dict):
    """Check for restart policy input

    :param container_dict: a container within the request body.
    """
    restart_policy = container_dict.get('restart_policy')
    if not restart_policy:
        return

    name = restart_policy.get('Name')
    num = restart_policy.setdefault('MaximumRetryCount', '0')
    count = int(num)
    if name in ['unless-stopped', 'always']:
        if count != 0:
            msg = _("maximum retry count not valid with restart "
                    "policy of %s") % name
            raise exception.InvalidValue(msg)
    elif name in ['no']:
        container_dict.get('restart_policy')['MaximumRetryCount'] = '0'


def build_exposed_ports(ports):

    def validate_protocol(protocol):
        if protocol not in ('tcp', 'udp'):
            raise exception.InvalidValue(_(
                "value %s is an invalid protocol") % protocol)

    def validate_port(port):
        try:
            int(port)
        except ValueError:
            msg = _("value %s is invalid as publish port.") % port
            raise exception.InvalidValue(msg)

    exposed_ports = {}
    for key, value in ports.items():
        try:
            port, protocol = key.split('/')
        except ValueError:
            port, protocol = key, 'tcp'

        validate_protocol(protocol)
        validate_port(port)

        key = '/'.join([port, protocol])
        exposed_ports[key] = value

    return exposed_ports


def build_requested_networks(context, nets):
    """Build requested networks by calling neutron client

    :param nets: The special network uuid when create container
                 if none, will call neutron to create new network.
    :returns: available network and ports
    """
    neutron_api = neutron.NeutronAPI(context)
    requested_networks = []
    for net in nets:
        if net.get('port'):
            port = neutron_api.get_neutron_port(net['port'])
            neutron_api.ensure_neutron_port_usable(port)
            network = neutron_api.get_neutron_network(port['network_id'])
            requested_networks.append({'network': port['network_id'],
                                       'port': port['id'],
                                       'router:external':
                                           network.get('router:external'),
                                       'shared': network.get('shared'),
                                       'fixed_ip': '',
                                       'preserve_on_delete': True})
        elif net.get('network'):
            network = neutron_api.get_neutron_network(net['network'])
            requested_networks.append({'network': network['id'],
                                       'port': '',
                                       'router:external':
                                           network.get('router:external'),
                                       'shared': network.get('shared'),
                                       'fixed_ip':
                                           net.get('fixed_ip') or
                                           net.get('v4-fixed-ip', '') or
                                           net.get('v6-fixed-ip', ''),
                                       'preserve_on_delete': False})
    if not requested_networks:
        # Find an available neutron net and create docker network by
        # wrapping the neutron net.
        network = neutron_api.get_available_network()
        requested_networks.append({'network': network['id'],
                                   'port': '',
                                   'router:external':
                                       network.get('router:external'),
                                   'shared': network.get('shared'),
                                   'fixed_ip': '',
                                   'preserve_on_delete': False})

    check_external_network_attach(context, requested_networks)
    return requested_networks


def check_external_network_attach(context, nets):
    """Check if attaching to external network is permitted."""
    if not context.can(NETWORK_ATTACH_EXTERNAL,
                       fatal=False):
        for net in nets:
            if net.get('router:external') and not net.get('shared'):
                raise exception.ExternalNetworkAttachForbidden(
                    network_uuid=net['network'])


class EventReporter(object):
    """Context manager to report container action events."""

    def __init__(self, context, event_name, *container_uuids):
        self.context = context
        self.event_name = event_name
        self.container_uuids = container_uuids

    def __enter__(self):
        for uuid in self.container_uuids:
            objects.ContainerActionEvent.event_start(
                self.context, uuid, self.event_name, want_result=False)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for uuid in self.container_uuids:
            objects.ContainerActionEvent.event_finish(
                self.context, uuid, self.event_name, exc_val=exc_val,
                exc_tb=exc_tb, want_result=False)
        return False


class FinishAction(object):
    """Context manager to finish container actions."""

    def __init__(self, context, action_name, *container_uuids):
        self.context = context
        self.action_name = action_name
        self.container_uuids = container_uuids

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for uuid in self.container_uuids:
            objects.ContainerAction.action_finish(
                self.context, uuid, self.action_name, exc_val=exc_val,
                exc_tb=exc_tb, want_result=False)
        return False


def get_wrapped_function(function):
    """Get the method at the bottom of a stack of decorators."""
    if not hasattr(function, '__closure__') or not function.__closure__:
        return function

    def _get_wrapped_function(function):
        if not hasattr(function, '__closure__') or not function.__closure__:
            return None

        for closure in function.__closure__:
            func = closure.cell_contents

            deeper_func = _get_wrapped_function(func)
            if deeper_func:
                return deeper_func
            elif hasattr(closure.cell_contents, '__call__'):
                return closure.cell_contents

        return function

    return _get_wrapped_function(function)


def wrap_container_event(prefix, finish_action=None):
    """Warps a method to log the event taken on the container, and result.

    This decorator wraps a method to log the start and result of an event, as
    part of an action taken on a container.
    """
    def helper(function):

        @functools.wraps(function)
        def decorated_function(self, context, *args, **kwargs):
            wrapped_func = get_wrapped_function(function)
            keyed_args = inspect.getcallargs(wrapped_func, self, context,
                                             *args, **kwargs)
            container_uuid = keyed_args['container'].uuid
            event_name = '{0}_{1}'.format(prefix, function.__name__)

            if finish_action is not None:
                with FinishAction(
                    context, finish_action, container_uuid
                ), EventReporter(
                    context, event_name, container_uuid
                ):
                    return function(self, context, *args, **kwargs)
            else:
                with EventReporter(context, event_name, container_uuid):
                    return function(self, context, *args, **kwargs)
        return decorated_function
    return helper


def wrap_exception():
    def helper(function):

        @functools.wraps(function)
        def decorated_function(self, context, container, *args, **kwargs):
            try:
                return function(self, context, container, *args, **kwargs)
            except exception.DockerError as e:
                with excutils.save_and_reraise_exception(reraise=False):
                    LOG.error("Error occurred while calling Docker API: %s",
                              six.text_type(e))
            except Exception as e:
                with excutils.save_and_reraise_exception(reraise=False):
                    LOG.exception("Unexpected exception: %s", six.text_type(e))
        return decorated_function
    return helper


def is_close(x, y, rel_tol=1e-06, abs_tol=0.0):
    return abs(x - y) <= max(rel_tol * max(abs(x), abs(y)), abs_tol)


def is_less_than(x, y):
    if isinstance(x, int) and isinstance(y, int):
        return x < y
    if isinstance(x, float) or isinstance(y, float):
        return False if (x - y) >= 0 or is_close(x, y) else True


def encode_file_data(data):
    if six.PY3 and isinstance(data, str):
        data = data.encode('utf-8')
    return base64.b64encode(data).decode('utf-8')


def decode_file_data(data):
    # Py3 raises binascii.Error instead of TypeError as in Py27
    try:
        return base64.b64decode(data)
    except (TypeError, binascii.Error):
        raise exception.Base64Exception()


def strtime(at):
    return at.strftime("%Y-%m-%dT%H:%M:%S.%f")


if six.PY2:
    nested_contexts = contextlib.nested
else:
    @contextlib.contextmanager
    def nested_contexts(*contexts):
        with contextlib.ExitStack() as stack:
            yield [stack.enter_context(c) for c in contexts]


def convert_mb_to_ceil_gb(mb_value):
    gb_int = 0
    if mb_value:
        gb_float = mb_value / 1024.0
        # ensure we reserve/allocate enough space by rounding up to nearest GB
        gb_int = int(math.ceil(gb_float))
    return gb_int


if hasattr(inspect, 'getfullargspec'):
    getargspec = inspect.getfullargspec
else:
    getargspec = inspect.getargspec


def expects_func_args(*args):
    def _decorator_checker(dec):
        @functools.wraps(dec)
        def _decorator(f):
            base_f = get_wrapped_function(f)
            argspec = getargspec(base_f)
            if argspec[1] or argspec[2] or set(args) <= set(argspec[0]):
                # NOTE (ndipanov): We can't really tell if correct stuff will
                # be passed if it's a function with *args or **kwargs so
                # we still carry on and hope for the best
                return dec(f)
            else:
                raise TypeError("Decorated function %(f_name)s does not "
                                "have the arguments expected by the "
                                "decorator %(d_name)s" %
                                {'f_name': base_f.__name__,
                                 'd_name': dec.__name__})
        return _decorator
    return _decorator_checker



# Copyright 2017 IBM Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_concurrency import processutils
from oslo_serialization import jsonutils
import six

from zun.common import exception
from zun.common import utils
from zun import objects
from zun.objects import fields
from zun.pci import utils as pci_utils


class Host(object):

    def __init__(self):
        self.capabilities = None

    def get_cpu_numa_info(self):
        """This method returns a dict containing the cpuset info for a host"""

        raise NotImplementedError()

    def get_host_numa_topology(self, numa_topo_obj):
        # Replace this call with a more generic call when we obtain other
        # NUMA related data like memory etc.
        cpu_info = self.get_cpu_numa_info()
        mem_info = self.get_mem_numa_info()
        floating_cpus = utils.get_floating_cpu_set()
        numa_node_obj = []
        for cpu, mem_total in zip(cpu_info.items(), mem_info):
            node, cpuset = cpu
            numa_node = objects.NUMANode()
            if floating_cpus:
                allowed_cpus = set(cpuset) - (floating_cpus & set(cpuset))
            else:
                allowed_cpus = set(cpuset)
            numa_node.id = node
            # allowed_cpus are the ones allowed to pin on.
            # Rest of the cpus are assumed to be floating
            # in nature.
            numa_node.cpuset = allowed_cpus
            numa_node.pinned_cpus = set([])
            numa_node.mem_total = mem_total
            numa_node.mem_available = mem_total
            numa_node_obj.append(numa_node)
        numa_topo_obj.nodes = numa_node_obj

    def get_host_mem(self):
        with open('/proc/meminfo') as fp:
            m = fp.read().split()
            idx1 = m.index('MemTotal:')
            mem_total = m[idx1 + 1]
            idx2 = m.index('MemFree:')
            mem_free = m[idx2 + 1]
            # MemAvailable is not available until 3.14 kernel
            if 'MemAvailable:' in m:
                idx3 = m.index('MemAvailable:')
                mem_ava = m[idx3 + 1]
            else:
                idx31 = m.index('Buffers:')
                buffers = m[idx31 + 1]
                idx32 = m.index('Cached:')
                cached = m[idx32 + 1]
                mem_ava = int(mem_free) + int(buffers) + int(cached)
            mem_used = int(mem_total) - int(mem_ava)
        return int(mem_total), int(mem_free), int(mem_ava), int(mem_used)

    def get_pci_resources(self):
        addresses = []
        try:
            output, status = utils.execute('lspci', '-D', '-nnmm')
            lines = output.split('\n')
            for line in lines:
                if not line:
                    continue
                columns = line.split()
                address = columns[0]
                addresses.append(address)
        except processutils.ProcessExecutionError as e:
            raise exception.CommandError(cmd='lspci',
                                         error=six.text_type(e))

        pci_info = []
        for addr in addresses:
            pci_info.append(self._get_pci_dev_info(addr))

        return jsonutils.dumps(pci_info)

    def _get_pci_dev_info(self, address):
        """Returns a dict of PCI device."""

        def _get_device_type(address):
            """Get a PCI device's device type.

            An assignable PCI device can be a normal PCI device,
            a SR-IOV Physical Function (PF), or a SR-IOV Virtual
            Function (VF). Only normal PCI devices or SR-IOV VFs
            are assignable.
            """
            path = '/sys/bus/pci/devices/' + address + '/'
            output, status = utils.execute('ls', path)
            if "physfn" in output:
                phys_address = None
                upath = '/sys/bus/pci/devices/%s/physfn/uevent' % address
                ou, st = utils.execute('cat', upath)
                lines = ou.split('\n')
                for line in lines:
                    if 'PCI_SLOT_NAME' in line:
                        columns = line.split("=")
                        phys_address = columns[1]
                return {'dev_type': fields.PciDeviceType.SRIOV_VF,
                        'parent_addr': phys_address}
            if "virtfn" in output:
                return {'dev_type': fields.PciDeviceType.SRIOV_PF}
            return {'dev_type': fields.PciDeviceType.STANDARD}

        def _get_device_capabilities(device, address):
            """Get PCI VF device's additional capabilities.

            If a PCI device is a virtual function, this function reads the PCI
            parent's network capabilities (must be always a NIC device) and
            appends this information to the device's dictionary.
            """
            if device.get('dev_type') == fields.PciDeviceType.SRIOV_VF:
                pcinet_info = self._get_pcinet_info(address)
                if pcinet_info:
                    return {'capabilities':
                            {'network': pcinet_info.get('capabilities')}}
            return {}

        def _get_vendor_and_product(address):
            output, status = utils.execute('lspci', '-n', '-s', address)
            value = output.split()[2]
            result = value.split(":")
            return result[0], result[1]

        def _get_numa_node(address):
            numa_node = None
            output, status = utils.execute('lspci', '-vmm', '-s', address)
            lines = output.split('\n')
            for line in lines:
                if 'NUMANode' in line:
                    numa_node = int(line.split(":")[1])
            return numa_node

        dev_name = 'pci_' + address.replace(":", "_").replace(".", "_")
        vendor_id, product_id = _get_vendor_and_product(address)
        numa_node = _get_numa_node(address)
        device = {
            "dev_id": dev_name,
            "address": address,
            "product_id": product_id,
            "vendor_id": vendor_id,
            "numa_node": numa_node
        }
        device['label'] = 'label_%(vendor_id)s_%(product_id)s' % device
        device.update(_get_device_type(address))
        device.update(_get_device_capabilities(device, address))
        return device

    def _get_pcinet_info(self, vf_address):
        """Returns a dict of NET device."""
        devname = pci_utils.get_net_name_by_vf_pci_address(vf_address)
        if not devname:
            return

        ifname = pci_utils.get_ifname_by_pci_address(vf_address)
        # Features from the that libvirt supported, get them by ethtool -k
        # Note: I cannot find the rdma feature returned by ethtool, correct me
        # if the string is wrong.
        FEATURES_LIST = ['rx-checksumming', 'tx-checksumming',
                         'scatter-gather', 'tcp-segmentation-offload',
                         'generic-segmentation-offload',
                         'generic-receive-offload', 'large-receive-offload',
                         'rx-vlan-offload', 'tx-vlan-offload',
                         'ntuple-filters', 'receive-hashing',
                         'tx-udp_tnl-segmentation', 'rdma']
        FEATURES_MAP = {'rx-checksumming': 'rx',
                        'tx-checksumming': 'tx',
                        'scatter-gather': 'sg',
                        'tcp-segmentation-offload': 'tso',
                        'generic-segmentation-offload': 'gso',
                        'generic-receive-offload': 'gro',
                        'large-receive-offload': 'lro',
                        'rx-vlan-offload': 'rxvlan',
                        'tx-vlan-offload': 'txvlan',
                        'ntuple-filters': 'ntuple',
                        'receive-hashing': 'rxhash',
                        'tx-udp_tnl-segmentation': 'txudptnl',
                        'rdma': 'rdma'}

        features = []
        output, status = utils.execute('ethtool', '-k', ifname)
        lines = output.split('\n')
        for line in lines:
            columns = line.split(":")
            if columns[0].strip() in FEATURES_LIST:
                if "on" in columns[1].strip():
                    features.append(FEATURES_MAP.get(columns[0].strip()))
        return {'name': devname,
                'capabilities': features}



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""add disk total and used to compute node

Revision ID: d0c606fdec3c
Revises: 3f49fa520409
Create Date: 2018-03-06 18:44:27.630273

"""

# revision identifiers, used by Alembic.
revision = 'd0c606fdec3c'
down_revision = '3f49fa520409'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    with op.batch_alter_table('compute_node', schema=None) as batch_op:
        batch_op.add_column(sa.Column('disk_total',
                                      sa.Integer(), nullable=False))
        batch_op.add_column(sa.Column('disk_used',
                                      sa.Integer(), nullable=False))



# Copyright (c) 2017 Intel, Inc.
# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

pci_group = cfg.OptGroup(
    name='pci',
    title='PCI passthrough options')

pci_opts = [
    cfg.MultiStrOpt('alias',
                    default=[],
                    help="""
An alias for a PCI passthrough device requirement.

Possible Values:

* A list of JSON values which describe the aliases. For example:

    alias = {
      "name": "QuickAssist",
      "product_id": "0443",
      "vendor_id": "8086",
      "device_type": "PCI"
    }

  defines an alias for the Intel QuickAssist card. (multi valued). Valid key
  values are :

  * "name": Name of the PCI alias.
  * "product_id": Product ID of the device in hexadecimal.
  * "vendor_id": Vendor ID of the device in hexadecimal.
  * "device_type": Type of PCI device. Valid values are: "type-PCI",
    "PF" and "VF".
"""),
    cfg.MultiStrOpt('passthrough_whitelist',
                    default=[],
                    help="""
White list of PCI devices available to VMs.

Possible values:

* A JSON dictionary which describe a whitelisted PCI device. It should take
  the following format:

    ["vendor_id": "<id>",] ["product_id": "<id>",]
    ["address": "[[[[<domain>]:]<bus>]:][<slot>][.[<function>]]" |
     "devname": "<name>",]
    {"<tag>": "<tag_value>",}

  Where '[' indicates zero or one occurrences, '{' indicates zero or multiple
  occurrences, and '|' mutually exclusive options. Note that any missing
  fields are automatically wildcarded.

  Valid key values are :

  * "vendor_id": Vendor ID of the device in hexadecimal.
  * "product_id": Product ID of the device in hexadecimal.
  * "address": PCI address of the device.
  * "devname": Device name of the device (for e.g. interface name). Not all
    PCI devices have a name.
  * "<tag>": Additional <tag> and <tag_value> used for matching PCI devices.
    Supported <tag>: "physical_network".

  The address key supports traditional glob style and regular expression
  syntax. Valid examples are:

    passthrough_whitelist = {"devname":"eth0",
                             "physical_network":"physnet"}
    passthrough_whitelist = {"address":"*:0a:00.*"}
    passthrough_whitelist = {"address":":0a:00.",
                             "physical_network":"physnet1"}
    passthrough_whitelist = {"vendor_id":"1137",
                             "product_id":"0071"}
    passthrough_whitelist = {"vendor_id":"1137",
                             "product_id":"0071",
                             "address": "0000:0a:00.1",
                             "physical_network":"physnet1"}
    passthrough_whitelist = {"address":{"domain": ".*",
                                        "bus": "02", "slot": "01",
                                        "function": "[2-7]"},
                             "physical_network":"physnet1"}
    passthrough_whitelist = {"address":{"domain": ".*",
                                        "bus": "02", "slot": "0[1-2]",
                                        "function": ".*"},
                             "physical_network":"physnet1"}

  The following are invalid, as they specify mutually exclusive options:

    passthrough_whitelist = {"devname":"eth0",
                             "physical_network":"physnet",
                             "address":"*:0a:00.*"}

* A JSON list of JSON dictionaries corresponding to the above format. For
  example:

    passthrough_whitelist = [{"product_id":"0001", "vendor_id":"8086"},
                             {"product_id":"0002", "vendor_id":"8086"}]
""")
]


def register_opts(conf):
    conf.register_group(pci_group)
    conf.register_opts(pci_opts, group=pci_group)


def list_opts():
    return {pci_group: pci_opts}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from __future__ import with_statement
from alembic import context
from logging.config import fileConfig

from zun.db.sqlalchemy import api as sqla_api
from zun.db.sqlalchemy import models
# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
target_metadata = models.Base.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    engine = sqla_api.get_engine()
    with engine.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            render_as_batch=True
        )

        with context.begin_transaction():
            context.run_migrations()

run_migrations_online()



#    Copyright 2016 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg


compute_opts = [
    cfg.BoolOpt(
        'resume_container_state',
        default=True,
        help='restart the containers which are running '
             'before the host reboots.'),
    cfg.FloatOpt(
        'reserve_disk_for_image',
        default=0.2,
        help='reserve disk for docker images'),
    cfg.BoolOpt(
        'enable_cpu_pinning',
        default=False,
        help='allow the container with cpu_policy is dedicated'),
    cfg.IntOpt(
        'resource_provider_association_refresh',
        default=300,
        min=0,
        mutable=True,
        # TODO(efried): Provide more/better explanation of what this option is
        # all about. Reference bug(s). Unless we're just going to remove it.
        help="""
Interval for updating zun-compute-side cache of the compute node resource
provider's inventories, aggregates, and traits.
This option specifies the number of seconds between attempts to update a
provider's inventories, aggregates and traits in the local cache of the compute
node.
A value of zero disables cache refresh completely.
The cache can be cleared manually at any time by sending SIGHUP to the compute
process, causing it to be repopulated the next time the data is accessed.
Possible values:
* Any positive integer in seconds, or zero to disable refresh.
"""),
    cfg.BoolOpt(
        'host_shared_with_nova',
        default=False,
        help='Whether this compute node is shared with nova'),
]

service_opts = [
    cfg.StrOpt(
        'topic',
        default='zun-compute',
        help='The queue to add compute tasks to.'),
]

db_opts = [
    cfg.StrOpt(
        'unique_container_name_scope',
        default='',
        choices=['', 'project', 'global'],
        help="""
Sets the scope of the check for unique container names.
The default doesn't check for unique names. If a scope for the name check is
set, a launch of a new container with a duplicate name will result in an
''ContainerAlreadyExists'' error. The uniqueness is case-insensitive.
Setting this option can increase the usability for end users as they don't
have to distinguish among containers with the same name by their IDs.
Possible values:
* '': An empty value means that no uniqueness check is done and duplicate
  names are possible.
* "project": The container name check is done only for containers within the
  same project.
* "global": The container name check is done for all containers regardless of
  the project.
"""),
]

resource_tracker_opts = [
    cfg.IntOpt('reserved_host_disk_mb',
               min=0,
               default=0,
               help="""
Amount of disk resources in MB to make them always available to host. The
disk usage gets reported back to the scheduler from zun-compute running
on the compute nodes. To prevent the disk resources from being considered
as available, this option can be used to reserve disk space for that host.
Possible values:
* Any positive integer representing amount of disk in MB to reserve
  for the host.
"""),
    cfg.IntOpt('reserved_host_memory_mb',
               default=512,
               min=0,
               help="""
Amount of memory in MB to reserve for the host so that it is always available
to host processes. The host resources usage is reported back to the scheduler
continuously from zun-compute running on the compute node. To prevent the host
memory from being considered as available, this option is used to reserve
memory for the host.
Possible values:
* Any positive integer representing amount of memory in MB to reserve
  for the host.
"""),
    cfg.IntOpt('reserved_host_cpus',
               default=0,
               min=0,
               help="""
Number of physical CPUs to reserve for the host. The host resources usage is
reported back to the scheduler continuously from zun-compute running on the
compute node. To prevent the host CPU from being considered as available,
this option is used to reserve random pCPU(s) for the host.
Possible values:
* Any positive integer representing number of physical CPUs to reserve
  for the host.
"""),
]

allocation_ratio_opts = [
    cfg.FloatOpt('cpu_allocation_ratio',
                 default=None,
                 min=0.0,
                 help="""
Virtual CPU to physical CPU allocation ratio.
This option is used to influence the hosts selected by the Placement API. In
addition, the ``AggregateCoreFilter`` will fall back to this configuration
value if no per-aggregate setting is found.
.. note::
   If this option is set to something *other than* ``None`` or ``0.0``, the
   allocation ratio will be overwritten by the value of this option, otherwise,
   the allocation ratio will not change. Once set to a non-default value, it is
   not possible to "unset" the config to get back to the default behavior. If
   you want to reset back to the initial value, explicitly specify it to the
   value of ``initial_cpu_allocation_ratio``.
Possible values:
* Any valid positive integer or float value
Related options:
* ``initial_cpu_allocation_ratio``
"""),
    cfg.FloatOpt('ram_allocation_ratio',
                 default=None,
                 min=0.0,
                 help="""
Virtual RAM to physical RAM allocation ratio.
This option is used to influence the hosts selected by the Placement API. In
addition, the ``AggregateRamFilter`` will fall back to this configuration value
if no per-aggregate setting is found.
.. note::
   If this option is set to something *other than* ``None`` or ``0.0``, the
   allocation ratio will be overwritten by the value of this option, otherwise,
   the allocation ratio will not change. Once set to a non-default value, it is
   not possible to "unset" the config to get back to the default behavior. If
   you want to reset back to the initial value, explicitly specify it to the
   value of ``initial_ram_allocation_ratio``.
Possible values:
* Any valid positive integer or float value
Related options:
* ``initial_ram_allocation_ratio``
"""),
    cfg.FloatOpt('disk_allocation_ratio',
                 default=None,
                 min=0.0,
                 help="""
Virtual disk to physical disk allocation ratio.
This option is used to influence the hosts selected by the Placement API. In
addition, the ``AggregateDiskFilter`` will fall back to this configuration
value if no per-aggregate setting is found.
When configured, a ratio greater than 1.0 will result in over-subscription of
the available physical disk, which can be useful for more efficiently packing
containers created with images that do not use the entire virtual disk.
It can be set to a value between 0.0 and 1.0 in
order to preserve a percentage of the disk for uses other than containers.
.. note::
   If the value is set to ``>1``, we recommend keeping track of the free disk
   space, as the value approaching ``0`` may result in the incorrect
   functioning of instances using it at the moment.
.. note::
   If this option is set to something *other than* ``None`` or ``0.0``, the
   allocation ratio will be overwritten by the value of this option, otherwise,
   the allocation ratio will not change. Once set to a non-default value, it is
   not possible to "unset" the config to get back to the default behavior. If
   you want to reset back to the initial value, explicitly specify it to the
   value of ``initial_disk_allocation_ratio``.
Possible values:
* Any valid positive integer or float value
Related options:
* ``initial_disk_allocation_ratio``
"""),
    cfg.FloatOpt('initial_cpu_allocation_ratio',
                 default=16.0,
                 min=0.0,
                 help="""
Initial virtual CPU to physical CPU allocation ratio.
This is only used when initially creating the ``computes_nodes`` table record
for a given zun-compute service.
Related options:
* ``cpu_allocation_ratio``
"""),
    cfg.FloatOpt('initial_ram_allocation_ratio',
                 default=1.5,
                 min=0.0,
                 help="""
Initial virtual RAM to physical RAM allocation ratio.
This is only used when initially creating the ``computes_nodes`` table record
for a given zun-compute service.
Related options:
* ``ram_allocation_ratio``
"""),
    cfg.FloatOpt('initial_disk_allocation_ratio',
                 default=1.0,
                 min=0.0,
                 help="""
Initial virtual disk to physical disk allocation ratio.
This is only used when initially creating the ``computes_nodes`` table record
for a given zun-compute service.
Related options:
* ``disk_allocation_ratio``
""")
]

opt_group = cfg.OptGroup(
    name='compute', title='Options for the zun-compute service')

ALL_OPTS = (service_opts + db_opts + compute_opts + resource_tracker_opts +
            allocation_ratio_opts)


def register_opts(conf):
    conf.register_group(opt_group)
    conf.register_opts(ALL_OPTS, opt_group)


def list_opts():
    return {opt_group: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock
from oslo_config import cfg

from zun.api import servicegroup
from zun import objects
from zun.tests.unit.api import base as api_base
from zun.tests.unit.api import utils as api_utils


class DbRec(object):

    def __init__(self, d):
        self.rec_as_dict = d

    def as_dict(self):
        return self.rec_as_dict


class TestZunServiceController(api_base.FunctionalTest):

    @mock.patch('zun.common.policy.enforce')
    def test_empty(self, mock_policy):
        mock_policy.return_value = True
        response = self.get_json('/services')
        self.assertEqual([], response['services'])

    def _rpc_api_reply(self, count=1, **kwarg):
        reclist = []
        for i in range(count):
            elem = api_utils.zservice_get_data(**kwarg)
            elem['id'] = i + 1
            rec = DbRec(elem)
            reclist.append(rec)
        return reclist

    @mock.patch('zun.common.policy.enforce')
    @mock.patch.object(objects.ZunService, 'list')
    @mock.patch.object(servicegroup.ServiceGroup, 'service_is_up')
    def test_get_one(self, svc_up, mock_list, mock_policy):
        mock_policy.return_value = True
        mock_list.return_value = self._rpc_api_reply()
        svc_up.return_value = "up"

        response = self.get_json('/services')
        self.assertEqual(1, len(response['services']))
        self.assertEqual(1, response['services'][0]['id'])
        self.assertEqual('fake-zone',
                         response['services'][0]['availability_zone'])

    @mock.patch('zun.common.policy.enforce')
    @mock.patch.object(objects.ZunService, 'list')
    @mock.patch.object(servicegroup.ServiceGroup, 'service_is_up')
    def test_get_many(self, svc_up, mock_list, mock_policy):
        mock_policy.return_value = True
        svc_num = 5
        mock_list.return_value = self._rpc_api_reply(svc_num)
        svc_up.return_value = "up"

        response = self.get_json('/services')
        self.assertEqual(len(response['services']), svc_num)
        for i in range(svc_num):
            elem = response['services'][i]
            self.assertEqual(elem['id'], i + 1)

    @mock.patch('zun.common.policy.enforce')
    @mock.patch.object(objects.ZunService, 'list')
    @mock.patch.object(servicegroup.ServiceGroup, 'service_is_up')
    def test_default_availability_zone(self, svc_up, mock_list, mock_policy):
        cfg.CONF.set_override("default_availability_zone", "default-zone")
        mock_policy.return_value = True
        mock_list.return_value = self._rpc_api_reply(availability_zone=None)
        svc_up.return_value = "up"

        response = self.get_json('/services')
        self.assertEqual(1, len(response['services']))
        self.assertEqual(1, response['services'][0]['id'])
        self.assertEqual('default-zone',
                         response['services'][0]['availability_zone'])

    @mock.patch('zun.common.policy.enforce')
    @mock.patch.object(objects.ZunService, 'get_by_host_and_binary')
    @mock.patch.object(objects.ZunService, 'update')
    def test_enable(self, mock_update, mock_get_host, mock_policy):
        mock_policy.return_value = True
        return_value = {
            'service': {
                'host': 'fake-host',
                'binary': 'fake-binary',
                'disabled': False,
                'disabled_reason': None
            },
        }
        params = {'binary': 'fake-binary', 'host': 'fake-host'}
        response = self.put_json('/services/enable', params)
        self.assertFalse(response.json['service']['disabled'])
        self.assertEqual(return_value, response.json)

    @mock.patch('zun.common.policy.enforce')
    @mock.patch.object(objects.ZunService, 'get_by_host_and_binary')
    @mock.patch.object(objects.ZunService, 'update')
    def test_disable(self, mock_update, mock_get_host, mock_policy):
        mock_policy.return_value = True
        return_value = {
            'service': {
                'host': 'fake-host',
                'binary': 'fake-binary',
                'disabled': True,
                'disabled_reason': 'abc'
            },
        }
        params = {'binary': 'fake-binary', 'host': 'fake-host',
                  'disabled_reason': 'abc'}
        response = self.put_json('/services/disable', params)
        self.assertTrue(response.json['service']['disabled'])
        self.assertEqual('abc', response.json['service']['disabled_reason'])
        self.assertEqual(return_value, response.json)

    @mock.patch('zun.common.policy.enforce')
    @mock.patch.object(objects.ZunService, 'get_by_host_and_binary')
    @mock.patch.object(objects.ZunService, 'update')
    def test_force_down(self, mock_force_down, mock_get_host, mock_policy):
        mock_policy.return_value = True
        return_value = {
            'service': {
                'host': 'fake-host',
                'binary': 'fake-binary',
                'forced_down': True
            },
        }
        params = {'binary': 'fake-binary', 'host': 'fake-host',
                  'forced_down': True}
        response = self.put_json('/services/force_down', params)
        self.assertTrue(response.json['service']['forced_down'])
        self.assertEqual(return_value, response.json)


class TestZunServiceEnforcement(api_base.FunctionalTest):

    def _common_policy_check(self, rule, func, *arg, **kwarg):
        self.policy.set_rules({rule: 'project:non_fake'})
        response = func(*arg, **kwarg)
        self.assertEqual(403, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(
            "Policy doesn't allow %s to be performed." % rule,
            response.json['errors'][0]['detail'])

    def test_policy_disallow_get_all(self):
        self._common_policy_check(
            'zun-service:get_all', self.get_json,
            '/services', expect_errors=True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg


volume_group = cfg.OptGroup(name='volume',
                            title='Options for the container volume')

volume_opts = [
    cfg.StrOpt('driver',
               default='cinder',
               deprecated_for_removal=True,
               help='Defines which driver to use for container volume.'),
    cfg.ListOpt('driver_list',
                default=['cinder', 'local'],
                help="""Defines the list of volume driver to use.
Possible values:
* ``cinder``
* ``local``
Services which consume this:
* ``zun-compute``
Interdependencies to other options:
* None
"""),
    cfg.StrOpt('volume_dir',
               default='$state_path/mnt',
               help='At which the docker volume will create.'),
    cfg.StrOpt('fstype',
               default='ext4',
               help='Default filesystem type for volume.'),
    cfg.BoolOpt('use_multipath',
                default=False,
                help="""
Use multipath connection of volume

Volumes can be connected as multipath devices. This will provide high
availability and fault tolerance.
"""),
]


ALL_OPTS = (volume_opts)


def register_opts(conf):
    conf.register_group(volume_group)
    conf.register_opts(ALL_OPTS, group=volume_group)


def list_opts():
    return {volume_group: ALL_OPTS}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from zun.common import consts
from zun import objects
from zun.tests.unit.db import base
from zun.tests.unit.db import utils


class TestQuotaClassObject(base.DbTestCase):

    def setUp(self):
        super(TestQuotaClassObject, self).setUp()
        self.fake_quota_class = utils.get_test_quota_class()

    def test_get_quota_class(self):
        class_name = self.fake_quota_class['class_name']
        resource = self.fake_quota_class['resource']
        with mock.patch.object(self.dbapi, 'quota_class_get',
                               autospec=True) as mock_get:
            mock_get.return_value = self.fake_quota_class
            quota_class = objects.QuotaClass.get(
                self.context, class_name, resource)
            mock_get.assert_called_once_with(
                self.context, class_name, resource)
            self.assertEqual(self.context, quota_class._context)

    def test_get_all_with_default(self):
        class_name = consts.DEFAULT_QUOTA_CLASS_NAME
        with mock.patch.object(self.dbapi, 'quota_class_get_default',
                               autospec=True) as mock_get_all:
            mock_get_all.return_value = {
                'class_name': class_name,
                'resource_1': 10,
                'resource_2': 20
            }
            quota_class_dict = objects.QuotaClass.get_all(self.context)
            mock_get_all.assert_called_once_with(self.context)
            self.assertEqual(class_name, quota_class_dict['class_name'])

    def test_get_all_with_class_name(self):
        class_name = self.fake_quota_class['class_name']
        with mock.patch.object(self.dbapi, 'quota_class_get_all_by_name',
                               autospec=True) as mock_get_all:
            mock_get_all.return_value = {
                'class_name': class_name,
                'resource_1': 10,
                'resource_2': 20
            }
            quota_class_dict = objects.QuotaClass.get_all(
                self.context, class_name)
            mock_get_all.assert_called_once_with(self.context, class_name)
            self.assertEqual(class_name, quota_class_dict['class_name'])

    def test_create_quota_class(self):
        class_name = self.fake_quota_class['class_name']
        resource = self.fake_quota_class['resource']
        hard_limit = self.fake_quota_class['hard_limit']
        with mock.patch.object(self.dbapi, 'quota_class_create',
                               autospec=True) as mock_create:
            mock_create.return_value = self.fake_quota_class
            quota_class = objects.QuotaClass(
                self.context, **utils.get_test_quota_class_value())
            quota_class.create(self.context)
            mock_create.assert_called_once_with(
                self.context, class_name, resource, hard_limit)

    def test_update_quota(self):
        class_name = self.fake_quota_class['class_name']
        resource = self.fake_quota_class['resource']
        with mock.patch.object(self.dbapi, 'quota_class_get',
                               autospec=True) as mock_get:
            mock_get.return_value = self.fake_quota_class
            with mock.patch.object(self.dbapi, 'quota_class_update',
                                   autospec=True) as mock_update:
                quota_class = objects.QuotaClass.get(
                    self.context, class_name, resource)
                quota_class.hard_limit = 100
                quota_class.update()

                mock_get.assert_called_once_with(
                    self.context, class_name, resource)
                mock_update.assert_called_once_with(
                    None, class_name, resource, 100)
                self.assertEqual(self.context, quota_class._context)
                self.assertEqual(100, quota_class.hard_limit)



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from testtools.matchers import HasLength

from zun import objects
from zun.tests.unit.db import base
from zun.tests.unit.db import utils


class TestRegistryObject(base.DbTestCase):

    def setUp(self):
        super(TestRegistryObject, self).setUp()
        self.fake_cpuset = utils.get_cpuset_dict()
        self.fake_registry = utils.get_test_registry(
            cpuset=self.fake_cpuset, cpu_policy='dedicated')

    def test_get_by_uuid(self):
        uuid = self.fake_registry['uuid']
        with mock.patch.object(self.dbapi, 'get_registry_by_uuid',
                               autospec=True) as mock_get_registry:
            mock_get_registry.return_value = self.fake_registry
            registry = objects.Registry.get_by_uuid(self.context, uuid)
            mock_get_registry.assert_called_once_with(self.context, uuid)
            self.assertEqual(self.context, registry._context)

    def test_get_by_name(self):
        name = self.fake_registry['name']
        with mock.patch.object(self.dbapi, 'get_registry_by_name',
                               autospec=True) as mock_get_registry:
            mock_get_registry.return_value = self.fake_registry
            registry = objects.Registry.get_by_name(self.context, name)
            mock_get_registry.assert_called_once_with(self.context, name)
            self.assertEqual(self.context, registry._context)

    def test_list(self):
        with mock.patch.object(self.dbapi, 'list_registries',
                               autospec=True) as mock_get_list:
            mock_get_list.return_value = [self.fake_registry]
            registries = objects.Registry.list(self.context)
            self.assertEqual(1, mock_get_list.call_count)
            self.assertThat(registries, HasLength(1))
            self.assertIsInstance(registries[0], objects.Registry)
            self.assertEqual(self.context, registries[0]._context)

    def test_list_with_filters(self):
        with mock.patch.object(self.dbapi, 'list_registries',
                               autospec=True) as mock_get_list:
            mock_get_list.return_value = [self.fake_registry]
            filt = {'username': 'fake_username'}
            registries = objects.Registry.list(self.context,
                                               filters=filt)
            self.assertEqual(1, mock_get_list.call_count)
            self.assertThat(registries, HasLength(1))
            self.assertIsInstance(registries[0], objects.Registry)
            self.assertEqual(self.context, registries[0]._context)
            mock_get_list.assert_called_once_with(self.context,
                                                  filters=filt,
                                                  limit=None, marker=None,
                                                  sort_key=None, sort_dir=None)

    def test_create(self):
        with mock.patch.object(self.dbapi, 'create_registry',
                               autospec=True) as mock_create_registry:
            mock_create_registry.return_value = self.fake_registry
            registry_dict = dict(self.fake_registry)
            registry = objects.Registry(self.context, **registry_dict)
            registry.create(self.context)
            mock_create_registry.assert_called_once_with(self.context,
                                                         self.fake_registry)
            self.assertEqual(self.context, registry._context)

    def test_destroy(self):
        uuid = self.fake_registry['uuid']
        with mock.patch.object(self.dbapi, 'get_registry_by_uuid',
                               autospec=True) as mock_get_registry:
            mock_get_registry.return_value = self.fake_registry
            with mock.patch.object(self.dbapi, 'destroy_registry',
                                   autospec=True) as mock_destroy_registry:
                registry = objects.Registry.get_by_uuid(self.context, uuid)
                registry.destroy()
                mock_get_registry.assert_called_once_with(self.context, uuid)
                mock_destroy_registry.assert_called_once_with(None, uuid)
                self.assertEqual(self.context, registry._context)

    def test_save(self):
        uuid = self.fake_registry['uuid']
        with mock.patch.object(self.dbapi, 'get_registry_by_uuid',
                               autospec=True) as mock_get_registry:
            mock_get_registry.return_value = self.fake_registry
            with mock.patch.object(self.dbapi, 'update_registry',
                                   autospec=True) as mock_update_registry:
                registry = objects.Registry.get_by_uuid(self.context, uuid)
                registry.domain = 'testdomain.io'
                registry.username = 'testuesrname'
                registry.password = 'testpassword'
                registry.name = 'testname'
                registry.save()

                mock_get_registry.assert_called_once_with(self.context, uuid)
                mock_update_registry.assert_called_once_with(
                    None, uuid,
                    {'domain': 'testdomain.io',
                     'username': 'testuesrname',
                     'password': 'testpassword',
                     'name': 'testname'})
                self.assertEqual(self.context, registry._context)



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""split volume_mapping table

Revision ID: 33cdd98bb9b2
Revises: 35cb52c5553f
Create Date: 2018-09-22 22:24:35.745666

"""

# revision identifiers, used by Alembic.
revision = '33cdd98bb9b2'
down_revision = '35cb52c5553f'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
from sqlalchemy import dialects


def MediumText():
    return sa.Text().with_variant(dialects.mysql.MEDIUMTEXT(), 'mysql')


VOLUME_MAPPING_TABLE = sa.Table(
    'volume_mapping', sa.MetaData(),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('uuid', sa.String(36), primary_key=True, nullable=False),
    sa.Column('project_id', sa.String(length=255), nullable=True),
    sa.Column('user_id', sa.String(length=255), nullable=True),
    sa.Column('cinder_volume_id', sa.String(36), nullable=True),
    sa.Column('volume_provider', sa.String(36), nullable=False),
    sa.Column('connection_info', MediumText(), nullable=True),
    sa.Column('volume_id', sa.Integer(), nullable=True),
    sa.Column('host', sa.String(length=255), nullable=True),
    sa.Column('auto_remove', sa.Boolean(), default=False, nullable=True),
    sa.Column('contents', MediumText(), nullable=True))


VOLUME_TABLE = sa.Table(
    'volume', sa.MetaData(),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('id', sa.Integer(), primary_key=True, nullable=False),
    sa.Column('uuid', sa.String(36), nullable=False),
    sa.Column('project_id', sa.String(length=255), nullable=True),
    sa.Column('user_id', sa.String(length=255), nullable=True),
    sa.Column('cinder_volume_id', sa.String(36), nullable=True),
    sa.Column('volume_provider', sa.String(36), nullable=False),
    sa.Column('connection_info', MediumText(), nullable=True),
    sa.Column('host', sa.String(length=255), nullable=True),
    sa.Column('auto_remove', sa.Boolean(), default=False, nullable=True),
    sa.Column('contents', MediumText(), nullable=True))


def create_volume_table():
    op.create_table(
        'volume',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(36), nullable=False),
        sa.Column('project_id', sa.String(length=255), nullable=True),
        sa.Column('user_id', sa.String(length=255), nullable=True),
        sa.Column('cinder_volume_id', sa.String(36), nullable=True),
        sa.Column('volume_provider', sa.String(36), nullable=False),
        sa.Column('connection_info', MediumText(), nullable=True),
        sa.Column('host', sa.String(length=255), nullable=True),
        sa.Column('auto_remove', sa.Boolean(), default=False, nullable=True),
        sa.Column('contents', MediumText(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid', name='uniq_volume0uuid'),
    )


def update_existing_records():
    op.add_column('volume_mapping',
                  sa.Column('volume_id', sa.Integer(),
                            nullable=True))

    # perform data migration between tables
    session = sa.orm.Session(bind=op.get_bind())
    with session.begin(subtransactions=True):
        for row in session.query(VOLUME_MAPPING_TABLE):
            res = session.execute(
                VOLUME_TABLE.insert().values(
                    created_at=row.created_at,
                    updated_at=row.updated_at,
                    uuid=row.uuid,
                    project_id=row.project_id,
                    user_id=row.user_id,
                    cinder_volume_id=row.cinder_volume_id,
                    volume_provider=row.volume_provider,
                    connection_info=row.connection_info,
                    host=row.host)
            )
            session.execute(
                VOLUME_MAPPING_TABLE.update().values(
                    volume_id=res.inserted_primary_key[0]).where(
                        VOLUME_MAPPING_TABLE.c.id == row.id)
            )
    # this commit is necessary to allow further operations
    session.commit()

    op.alter_column('volume_mapping', 'volume_id',
                    nullable=False,
                    existing_type=sa.Integer(), existing_nullable=True,
                    existing_server_default=False)
    # add the constraint now that everything is populated on that table
    op.create_foreign_key(
        constraint_name=None, source_table='volume_mapping',
        referent_table='volume',
        local_cols=['volume_id'], remote_cols=['id'],
        ondelete='CASCADE')


def update_volume_mapping_table():
    op.drop_column('volume_mapping', 'cinder_volume_id')
    op.drop_column('volume_mapping', 'volume_provider')
    op.drop_column('volume_mapping', 'connection_info')
    op.drop_column('volume_mapping', 'contents')
    op.drop_column('volume_mapping', 'auto_remove')
    op.drop_column('volume_mapping', 'host')


def upgrade():
    create_volume_table()
    update_existing_records()
    update_volume_mapping_table()



#    Copyright 2017 Arm Limited.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from testtools.matchers import HasLength

from zun.common import consts
from zun import objects
from zun.tests.unit.db import base
from zun.tests.unit.db import utils


class TestCapsuleObject(base.DbTestCase):

    def setUp(self):
        super(TestCapsuleObject, self).setUp()
        self.fake_capsule = utils.get_test_container(
            container_type=consts.TYPE_CAPSULE)
        self.fake_capsule.pop('capsule_id')

    def test_get_by_uuid(self):
        uuid = self.fake_capsule['uuid']
        with mock.patch.object(self.dbapi, 'get_container_by_uuid',
                               autospec=True) as mock_get_capsule:
            mock_get_capsule.return_value = self.fake_capsule
            capsule = objects.Capsule.get_by_uuid(self.context, uuid)
            mock_get_capsule.assert_called_once_with(
                self.context, consts.TYPE_CAPSULE, uuid)
            self.assertEqual(self.context, capsule._context)

    def test_list(self):
        with mock.patch.object(self.dbapi, 'list_containers',
                               autospec=True) as mock_get_list:
            mock_get_list.return_value = [self.fake_capsule]
            capsules = objects.Capsule.list(self.context)
            self.assertEqual(1, mock_get_list.call_count)
            self.assertThat(capsules, HasLength(1))
            self.assertIsInstance(capsules[0], objects.Capsule)
            self.assertEqual(self.context, capsules[0]._context)

    def test_list_with_filters(self):
        with mock.patch.object(self.dbapi, 'list_containers',
                               autospec=True) as mock_get_list:
            mock_get_list.return_value = [self.fake_capsule]
            filt = {'status': 'Running'}
            capsules = objects.Capsule.list(self.context,
                                            filters=filt)
            self.assertEqual(1, mock_get_list.call_count)
            self.assertThat(capsules, HasLength(1))
            self.assertIsInstance(capsules[0], objects.Capsule)
            self.assertEqual(self.context, capsules[0]._context)
            mock_get_list.assert_called_once_with(self.context,
                                                  consts.TYPE_CAPSULE,
                                                  filters=filt,
                                                  limit=None, marker=None,
                                                  sort_key=None, sort_dir=None)

    def test_create(self):
        with mock.patch.object(self.dbapi, 'create_container',
                               autospec=True) as mock_create_capsule:
            mock_create_capsule.return_value = self.fake_capsule
            capsule = objects.Capsule(self.context, **self.fake_capsule)
            capsule.create(self.context)
            self.fake_capsule.pop('cpuset', None)
            mock_create_capsule.assert_called_once_with(self.context,
                                                        self.fake_capsule)
            self.assertEqual(self.context, capsule._context)

    def test_status_reason_in_fields(self):
        with mock.patch.object(self.dbapi, 'create_container',
                               autospec=True) as mock_create_capsule:
            mock_create_capsule.return_value = self.fake_capsule
            capsule = objects.Capsule(self.context, **self.fake_capsule)
            self.assertTrue(hasattr(capsule, 'status_reason'))
            capsule.status_reason = "Docker Error happened"
            capsule.create(self.context)
            self.assertEqual(
                "Docker Error happened",
                mock_create_capsule.call_args_list[0][0][1]['status_reason'])

    def test_destroy(self):
        uuid = self.fake_capsule['uuid']
        with mock.patch.object(self.dbapi, 'get_container_by_uuid',
                               autospec=True) as mock_get_capsule:
            mock_get_capsule.return_value = self.fake_capsule
            with mock.patch.object(self.dbapi, 'destroy_container',
                                   autospec=True) as mock_destroy_capsule:
                capsule = objects.Capsule.get_by_uuid(self.context, uuid)
                capsule.destroy()
                mock_get_capsule.assert_called_once_with(
                    self.context, consts.TYPE_CAPSULE, uuid)
                mock_destroy_capsule.assert_called_once_with(
                    None, consts.TYPE_CAPSULE, uuid)
                self.assertEqual(self.context, capsule._context)

    def test_save(self):
        uuid = self.fake_capsule['uuid']
        with mock.patch.object(self.dbapi, 'get_container_by_uuid',
                               autospec=True) as mock_get_capsule:
            mock_get_capsule.return_value = self.fake_capsule
            with mock.patch.object(self.dbapi, 'update_container',
                                   autospec=True) as mock_update_capsule:
                capsule = objects.Capsule.get_by_uuid(self.context, uuid)
                capsule.name = 'fake-meta-name-new'
                capsule.labels = {'key3': 'val3', 'key4': 'val4'}
                capsule.save()
                mock_get_capsule.assert_called_once_with(
                    self.context, consts.TYPE_CAPSULE, uuid)
                mock_update_capsule.assert_called_once_with(
                    None, consts.TYPE_CAPSULE, uuid,
                    {'name': 'fake-meta-name-new',
                     'labels': {'key3': 'val3', 'key4': 'val4'}})
                self.assertEqual(self.context, capsule._context)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Filter support
"""

from oslo_log import log as logging

from zun.scheduler import loadables

LOG = logging.getLogger(__name__)


class BaseFilter(object):
    """Base class for all filter classes."""
    def _filter_one(self, obj, container, extra_spec):
        """Return True if it passes the filter, False otherwise."""
        return True

    def filter_all(self, filter_obj_list, container, extra_spec):
        """Yield objects that pass the filter.

        Can be overridden in a subclass, if you need to base filtering
        decisions on all objects.  Otherwise, one can just override
        _filter_one() to filter a single object.
        """
        for obj in filter_obj_list:
            if self._filter_one(obj, container, extra_spec):
                yield obj

    # Set to true in a subclass if a filter only needs to be run once
    # for each request rather than for each instance
    run_filter_once_per_request = False

    def run_filter_for_index(self, index):
        """Return True or False,

        if the filter needs to be run for the "index-th" instance in a
        request, return True.  Only need to override this if a filter
        needs anything other than "first only" or "all" behaviour.
        """
        if self.run_filter_once_per_request and index > 0:
            return False
        else:
            return True


class BaseFilterHandler(loadables.BaseLoader):
    """Base class to handle loading filter classes.

    This class should be subclassed where one needs to use filters.
    """

    def get_filtered_objects(self, filters, objs, container, extra_spec,
                             index=0):
        list_objs = list(objs)
        LOG.debug("Starting with %d host(s)", len(list_objs))
        part_filter_results = []
        full_filter_results = []
        log_msg = "%(cls_name)s: (start: %(start)s, end: %(end)s)"
        for filter_ in filters:
            if filter_.run_filter_for_index(index):
                cls_name = filter_.__class__.__name__
                start_count = len(list_objs)
                objs = filter_.filter_all(list_objs, container, extra_spec)
                if objs is None:
                    LOG.debug("Filter %s says to stop filtering", cls_name)
                    return
                list_objs = list(objs)
                end_count = len(list_objs)
                part_filter_results.append(log_msg % {"cls_name": cls_name,
                                                      "start": start_count,
                                                      "end": end_count})
                if list_objs:
                    remaining = [(getattr(obj, "host", obj),
                                  getattr(obj, "nodename", ""))
                                 for obj in list_objs]
                    full_filter_results.append((cls_name, remaining))
                else:
                    LOG.info("Filter %s returned 0 hosts", cls_name)
                    full_filter_results.append((cls_name, None))
                    break
                LOG.debug("Filter %(cls_name)s returned "
                          "%(obj_len)d host(s)",
                          {'cls_name': cls_name, 'obj_len': len(list_objs)})
        if not list_objs:
            cnt_uuid = container.uuid
            msg_dict = {"cnt_uuid": cnt_uuid,
                        "str_results": str(full_filter_results)}
            full_msg = ("Filtering removed all hosts for the request with "
                        "container ID "
                        "'%(cnt_uuid)s'. Filter results: %(str_results)s"
                        ) % msg_dict
            msg_dict["str_results"] = str(part_filter_results)
            part_msg = ("Filtering removed all hosts for the request with "
                        "container ID "
                        "'%(cnt_uuid)s'. Filter results: %(str_results)s"
                        ) % msg_dict
            LOG.debug(full_msg)
            LOG.info(part_msg)
        return list_objs



# Copyright (c) 2013 dotCloud, Inc.
# Copyright 2014 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
A Docker Hypervisor which allows running Linux Containers instead of VMs.
"""

import os
import shutil
import six
import socket
import time

from docker import errors
import eventlet
from oslo_config import cfg
from oslo_log import log
from oslo_utils import fileutils
from oslo_utils import importutils
from oslo_utils import units
from oslo_utils import uuidutils
from oslo_utils import versionutils

from nova.compute import flavors
from nova.compute import power_state
from nova import exception
from nova.i18n import _
from nova import objects
from nova.objects import fields
from nova import utils
from nova.virt import driver
from nova.virt import firewall
from nova.virt import hardware
from nova.virt import hostutils
from nova.virt import images
from nova.virt.zun import client as docker_client
from nova.virt.zun import hostinfo
from nova.virt.zun import network


CONF = cfg.CONF
CONF.import_opt('instances_path', 'nova.compute.manager')

docker_opts = [
    cfg.StrOpt('root_directory',
               default='/var/lib/docker',
               help='Path to use as the root of the Docker runtime.'),
    cfg.StrOpt('host_url',
               default='unix:///var/run/docker.sock',
               help='tcp://host:port to bind/connect to or '
                    'unix://path/to/socket to use'),
    cfg.BoolOpt('api_insecure',
                default=False,
                help='If set, ignore any SSL validation issues'),
    cfg.StrOpt('ca_file',
               help='Location of CA certificates file for '
                    'securing docker api requests (tlscacert).'),
    cfg.StrOpt('cert_file',
               help='Location of TLS certificate file for '
                    'securing docker api requests (tlscert).'),
    cfg.StrOpt('key_file',
               help='Location of TLS private key file for '
                    'securing docker api requests (tlskey).'),
    cfg.StrOpt('vif_driver',
               default='nova.virt.zun.vifs.DockerGenericVIFDriver'),
    cfg.StrOpt('snapshots_directory',
               default='$instances_path/snapshots',
               help='Location where docker driver will temporarily store '
                    'snapshots.'),
    cfg.StrOpt('shared_directory',
               default=None,
               help='Shared directory where glance images located. If '
                    'specified, docker will try to load the image from '
                    'the shared directory by image ID.'),
    cfg.ListOpt('default_nameservers',
                default=['8.8.8.8', '8.8.4.4'],
                help='The default DNS server to use.'),
]

CONF.register_opts(docker_opts, 'docker')

LOG = log.getLogger(__name__)


class DockerDriver(driver.ComputeDriver):
    """Docker hypervisor driver."""

    def __init__(self, virtapi):
        super(DockerDriver, self).__init__(virtapi)
        self._docker = None
        vif_class = importutils.import_class(CONF.docker.vif_driver)
        self.vif_driver = vif_class()
        self.firewall_driver = firewall.load_driver(
            default='nova.virt.firewall.NoopFirewallDriver')
        # NOTE(zhangguoqing): For passing the nova unit tests
        self.active_migrations = {}

    @property
    def docker(self):
        if self._docker is None:
            self._docker = docker_client.DockerHTTPClient(CONF.docker.host_url)
        return self._docker

    def init_host(self, host):
        if self._is_daemon_running() is False:
            raise exception.NovaException(
                _('Docker daemon is not running or is not reachable'
                  ' (check the rights on /var/run/docker.sock)'))

    def _is_daemon_running(self):
        return self.docker.ping()

    def _start_firewall(self, instance, network_info):
        self.firewall_driver.setup_basic_filtering(instance, network_info)
        self.firewall_driver.prepare_instance_filter(instance, network_info)
        self.firewall_driver.apply_instance_filter(instance, network_info)

    def _stop_firewall(self, instance, network_info):
        self.firewall_driver.unfilter_instance(instance, network_info)

    def refresh_security_group_rules(self, security_group_id):
        """Refresh security group rules from data store.

        Invoked when security group rules are updated.

        :param security_group_id: The security group id.

        """
        self.firewall_driver.refresh_security_group_rules(security_group_id)

    def refresh_instance_security_rules(self, instance):
        """Refresh security group rules from data store.

        Gets called when an instance gets added to or removed from
        the security group the instance is a member of or if the
        group gains or loses a rule.

        :param instance: The instance object.

        """
        self.firewall_driver.refresh_instance_security_rules(instance)

    def ensure_filtering_rules_for_instance(self, instance, network_info):
        """Set up filtering rules.

        :param instance: The instance object.
        :param network_info: Instance network information.

        """
        self.firewall_driver.setup_basic_filtering(instance, network_info)
        self.firewall_driver.prepare_instance_filter(instance, network_info)

    def unfilter_instance(self, instance, network_info):
        """Stop filtering instance.

        :param instance: The instance object.
        :param network_info: Instance network information.

        """
        self.firewall_driver.unfilter_instance(instance, network_info)

    def list_instances(self, inspect=False):
        res = []
        for container in self.docker.containers(all=True):
            info = self.docker.inspect_container(container['id'])
            if not info:
                continue
            if inspect:
                res.append(info)
            else:
                res.append(info['Config'].get('Hostname'))
        return res

    def attach_interface(self, instance, image_meta, vif):
        """Attach an interface to the container."""
        self.vif_driver.plug(instance, vif)
        container_id = self._find_container_by_instance(instance).get('id')
        self.vif_driver.attach(instance, vif, container_id)

    def detach_interface(self, instance, vif):
        """Detach an interface from the container."""
        self.vif_driver.unplug(instance, vif)

    def plug_vifs(self, instance, network_info):
        """Plug VIFs into networks."""
        for vif in network_info:
            self.vif_driver.plug(instance, vif)
        self._start_firewall(instance, network_info)

    def _attach_vifs(self, instance, network_info):
        """Plug VIFs into container."""
        if not network_info:
            return

        container_id = self._get_container_id(instance)
        if not container_id:
            raise exception.InstanceNotFound(instance_id=instance['name'])
        netns_path = '/var/run/netns'
        if not os.path.exists(netns_path):
            utils.execute(
                'mkdir', '-p', netns_path, run_as_root=True)
        nspid = self._find_container_pid(container_id)
        if not nspid:
            msg = _('Cannot find any PID under container "{0}"')
            raise RuntimeError(msg.format(container_id))
        netns_path = os.path.join(netns_path, container_id)
        utils.execute(
            'ln', '-sf', '/proc/{0}/ns/net'.format(nspid),
            '/var/run/netns/{0}'.format(container_id),
            run_as_root=True)
        utils.execute('ip', 'netns', 'exec', container_id, 'ip', 'link',
                      'set', 'lo', 'up', run_as_root=True)

        for vif in network_info:
            self.vif_driver.attach(instance, vif, container_id)

    def unplug_vifs(self, instance, network_info):
        """Unplug VIFs from networks."""
        for vif in network_info:
            self.vif_driver.unplug(instance, vif)
        self._stop_firewall(instance, network_info)

    def _encode_utf8(self, value):
        if six.PY2 and not isinstance(value, six.text_type):
            value = six.text_type(value)
        return value.encode('utf-8')

    def _find_container_by_instance(self, instance):
        try:
            name = self._get_container_name(instance)
            containers = self.docker.containers(all=True,
                                                filters={'name': name})
            if containers:
                # NOTE(dims): We expect only one item in the containers list
                return self.docker.inspect_container(containers[0]['id'])
        except errors.APIError as e:
            if e.response.status_code != 404:
                raise
        return {}

    def _get_container_name(self, instance):
        return "zun-sandbox-" + instance['uuid']

    def _get_container_id(self, instance):
        return self._find_container_by_instance(instance).get('id')

    def get_info(self, instance):
        container = self._find_container_by_instance(instance)
        if not container:
            raise exception.InstanceNotFound(instance_id=instance['name'])
        running = container['State'].get('Running')
        mem = container['Config'].get('Memory', 0)

        # NOTE(ewindisch): cgroups/lxc defaults to 1024 multiplier.
        #                  see: _get_cpu_shares for further explanation
        num_cpu = container['Config'].get('CpuShares', 0) / 1024

        # FIXME(ewindisch): Improve use of statistics:
        #                   For 'mem', we should expose memory.stat.rss, and
        #                   for cpu_time we should expose cpuacct.stat.system,
        #                   but these aren't yet exposed by Docker.
        #
        #                   Also see:
        #                    docker/docs/sources/articles/runmetrics.md
        info = hardware.InstanceInfo(
            max_mem_kb=mem,
            mem_kb=mem,
            num_cpu=num_cpu,
            cpu_time_ns=0,
            state=(power_state.RUNNING if running
                   else power_state.SHUTDOWN)
        )
        return info

    def get_host_stats(self, refresh=False):
        hostname = socket.gethostname()
        stats = self.get_available_resource(hostname)
        stats['host_hostname'] = stats['hypervisor_hostname']
        stats['host_name_label'] = stats['hypervisor_hostname']
        return stats

    def get_available_nodes(self, refresh=False):
        hostname = socket.gethostname()
        return [hostname]

    def get_available_resource(self, nodename):
        if not hasattr(self, '_nodename'):
            self._nodename = nodename
        if nodename != self._nodename:
            LOG.error('Hostname has changed from %(old)s to %(new)s. '
                      'A restart is required to take effect.',
                      {'old': self._nodename, 'new': nodename})

        memory = hostinfo.get_memory_usage()
        disk = hostinfo.get_disk_usage()
        stats = {
            'vcpus': hostinfo.get_total_vcpus(),
            'vcpus_used': hostinfo.get_vcpus_used(self.list_instances(True)),
            'memory_mb': memory['total'] / units.Mi,
            'memory_mb_used': memory['used'] / units.Mi,
            'local_gb': disk['total'] / units.Gi,
            'local_gb_used': disk['used'] / units.Gi,
            'disk_available_least': disk['available'] / units.Gi,
            'hypervisor_type': 'docker',
            'hypervisor_version': versionutils.convert_version_to_int('1.0'),
            'hypervisor_hostname': self._nodename,
            'cpu_info': '?',
            'numa_topology': None,
            'supported_instances': [
                (fields.Architecture.I686, fields.HVType.DOCKER,
                 fields.VMMode.EXE),
                (fields.Architecture.X86_64, fields.HVType.DOCKER,
                 fields.VMMode.EXE)
            ]
        }
        return stats

    def _find_container_pid(self, container_id):
        n = 0
        while True:
            # NOTE(samalba): We wait for the process to be spawned inside the
            # container in order to get the "container pid". This is
            # usually really fast. To avoid race conditions on a slow
            # machine, we allow 10 seconds as a hard limit.
            if n > 20:
                return
            info = self.docker.inspect_container(container_id)
            if info:
                pid = info['State']['Pid']
                # Pid is equal to zero if it isn't assigned yet
                if pid:
                    return pid
            time.sleep(0.5)
            n += 1

    def _get_memory_limit_bytes(self, instance):
        if isinstance(instance, objects.Instance):
            return instance.get_flavor().memory_mb * units.Mi
        else:
            system_meta = utils.instance_sys_meta(instance)
            return int(system_meta.get(
                'instance_type_memory_mb', 0)) * units.Mi

    def _get_image_name(self, context, instance, image):
        fmt = image.container_format
        if fmt != 'docker':
            msg = _('Image container format not supported ({0})')
            raise exception.InstanceDeployFailure(msg.format(fmt),
                                                  instance_id=instance['name'])
        return image.name

    def _pull_missing_image(self, context, image_meta, instance):
        msg = 'Image name "%s" does not exist, fetching it...'
        LOG.debug(msg, image_meta.name)

        shared_directory = CONF.docker.shared_directory
        if (shared_directory and
                os.path.exists(os.path.join(shared_directory,
                                            image_meta.id))):
            LOG.debug('Found %s in shared_directory', image_meta.id)
            try:
                LOG.debug('Loading repository file into docker %s',
                          self._encode_utf8(image_meta.name))
                self.docker.load_repository_file(
                    self._encode_utf8(image_meta.name),
                    os.path.join(shared_directory, image_meta.id))
                return self.docker.inspect_image(
                    self._encode_utf8(image_meta.name))
            except Exception as e:
                # If failed to load image from shared_directory, continue
                # to download the image from glance then load.
                LOG.warning('Cannot load repository file from shared '
                            'directory: %s',
                            e, instance=instance, exc_info=True)

        # TODO(imain): It would be nice to do this with file like object
        # passing but that seems a bit complex right now.
        snapshot_directory = CONF.docker.snapshots_directory
        fileutils.ensure_tree(snapshot_directory)
        with utils.tempdir(dir=snapshot_directory) as tmpdir:
            try:
                out_path = os.path.join(tmpdir,
                                        uuidutils.generate_uuid(dashed=False))

                LOG.debug('Fetching image with id %s from glance',
                          image_meta.id)
                images.fetch(context, image_meta.id, out_path)
                LOG.debug('Loading repository file into docker %s',
                          self._encode_utf8(image_meta.name))
                self.docker.load_repository_file(
                    self._encode_utf8(image_meta.name),
                    out_path
                )
                return self.docker.inspect_image(
                    self._encode_utf8(image_meta.name))
            except Exception as e:
                LOG.warning('Cannot load repository file: %s',
                            e, instance=instance, exc_info=True)
                msg = _('Cannot load repository file: {0}')
                raise exception.NovaException(msg.format(e),
                                              instance_id=image_meta.name)

    def _create_instance_file(self, id, name, data):
        file_dir = os.path.join(CONF.instances_path, id)
        fileutils.ensure_tree(file_dir)
        file = os.path.join(file_dir, name)
        with open(file, 'a') as f:
            f.write(data)
        os.chmod(file_dir, 0o700)
        os.chmod(file, 0o600)
        return file

    def _cleanup_instance_file(self, id):
        dir = os.path.join(CONF.instances_path, id)
        if os.path.exists(dir):
            LOG.info('Deleting instance files %s', dir)
            try:
                shutil.rmtree(dir)
            except OSError as e:
                LOG.error(('Failed to cleanup directory %(target)s: '
                           '%(e)s'), {'target': dir, 'e': e})

    def _neutron_failed_callback(self, event_name, instance):
        LOG.error(('Neutron Reported failure on event '
                   '%(event)s for instance %(uuid)s'),
                  {'event': event_name, 'uuid': instance.uuid},
                  instance=instance)
        if CONF.vif_plugging_is_fatal:
            raise exception.VirtualInterfaceCreateException()

    def _get_neutron_events(self, network_info):
        # NOTE(danms): We need to collect any VIFs that are currently
        # down that we expect a down->up event for. Anything that is
        # already up will not undergo that transition, and for
        # anything that might be stale (cache-wise) assume it's
        # already up so we don't block on it.
        return [('network-vif-plugged', vif['id'])
                for vif in network_info if vif.get('active', True) is False]

    def _start_container(self, container_id, instance, network_info=None):
        self.docker.start(container_id)

        if not network_info:
            return
        timeout = CONF.vif_plugging_timeout
        if (utils.is_neutron() and timeout):
            events = self._get_neutron_events(network_info)
        else:
            events = []

        try:
            with self.virtapi.wait_for_instance_event(
                    instance, events, deadline=timeout,
                    error_callback=self._neutron_failed_callback):
                self.plug_vifs(instance, network_info)
                self._attach_vifs(instance, network_info)
        except eventlet.timeout.Timeout:
            LOG.warning(('Timeout waiting for vif plugging callback for '
                         'instance %(uuid)s'), {'uuid': instance['name']})
            if CONF.vif_plugging_is_fatal:
                self.docker.kill(container_id)
                self.docker.remove_container(container_id, force=True)
                raise exception.InstanceDeployFailure(
                    'Timeout waiting for vif plugging',
                    instance_id=instance['name'])
        except (Exception) as e:
            LOG.warning('Cannot setup network: %s',
                        e, instance=instance, exc_info=True)
            msg = _('Cannot setup network: {0}')
            self.docker.kill(container_id)
            self.docker.remove_container(container_id, force=True)
            raise exception.InstanceDeployFailure(msg.format(e),
                                                  instance_id=instance['name'])

    def spawn(self, context, instance, image_meta, injected_files,
              admin_password, network_info=None, block_device_info=None,
              flavor=None):
        image_name = self._get_image_name(context, instance, image_meta)
        args = {
            'hostname': instance['display_name'][:63],
            'mem_limit': self._get_memory_limit_bytes(instance),
            'cpu_shares': self._get_cpu_shares(instance),
            'network_disabled': True,
            'binds': self._get_binds(instance, network_info),
        }

        try:
            image = self.docker.inspect_image(self._encode_utf8(image_name))
        except errors.APIError:
            image = None

        if not image:
            image = self._pull_missing_image(context, image_meta, instance)

        container = self._create_container(instance, image_name, args)
        if not container:
            raise exception.InstanceDeployFailure(
                _('Cannot create container'),
                instance_id=instance['name'])

        container_id = container['Id']
        self._start_container(container_id, instance, network_info)

    def _get_binds(self, instance, network_info):
        binds = []
        dns = self._extract_dns_entries(network_info)
        bind = self._get_resolvconf_bind(instance['uuid'], dns)
        binds.append(bind)

        hostname = instance['display_name']
        bind = self._get_hostname_bind(instance['uuid'], hostname)
        binds.append(bind)

        bind = self._get_hosts_bind(instance['uuid'], hostname)
        binds.append(bind)
        return binds

    def _extract_dns_entries(self, network_info):
        dns = []
        if network_info:
            for net in network_info:
                subnets = net['network'].get('subnets', [])
                for subnet in subnets:
                    dns_entries = subnet.get('dns', [])
                    for dns_entry in dns_entries:
                        if 'address' in dns_entry:
                            dns.append(dns_entry['address'])
        return dns if dns else CONF.docker.default_nameservers

    def _get_resolvconf_bind(self, instance_id, nameservers):
        data = '\n'.join('nameserver %(server)s' % {'server': s}
                         for s in nameservers)
        file_name = 'resolv.conf'
        host_src = self._create_instance_file(instance_id, file_name, data)
        bind = '%(host_src)s:/etc/resolv.conf:ro' % {'host_src': host_src}
        return bind

    def _get_hostname_bind(self, instance_id, hostname):
        data = hostname
        file_name = 'hostname'
        host_src = self._create_instance_file(instance_id, file_name, data)
        bind = '%(host_src)s:/etc/hostname:ro' % {'host_src': host_src}
        return bind

    def _get_hosts_bind(self, instance_id, hostname):
        data = ('127.0.0.1 localhost ' + hostname + '\n'
                '::1     localhost ip6-localhost ip6-loopback\n'
                'fe00::0 ip6-localnet\n'
                'ff00::0 ip6-mcastprefix\n'
                'ff02::1 ip6-allnodes\n'
                'ff02::2 ip6-allrouters\n')
        file_name = 'hosts'
        host_src = self._create_instance_file(instance_id, file_name, data)
        bind = '%(host_src)s:/etc/hosts:ro' % {'host_src': host_src}
        return bind

    def restore(self, instance):
        container_id = self._get_container_id(instance)
        if not container_id:
            return

        self._start_container(container_id, instance)

    def _stop(self, container_id, instance, timeout=5):
        try:
            self.docker.stop(container_id, max(timeout, 5))
        except errors.APIError as e:
            if 'Unpause the container before stopping' not in e.explanation:
                LOG.warning('Cannot stop container: %s',
                            e, instance=instance, exc_info=True)
                raise
            self.docker.unpause(container_id)
            self.docker.stop(container_id, timeout)

    def soft_delete(self, instance):
        container_id = self._get_container_id(instance)
        if not container_id:
            return
        self._stop(container_id, instance)

    def destroy(self, context, instance, network_info, block_device_info=None,
                destroy_disks=True, migrate_data=None):
        self.soft_delete(instance)
        container_id = self._get_container_id(instance)
        if container_id:
            self.docker.remove_container(container_id, force=True)
        self.cleanup(context, instance, network_info,
                     block_device_info, destroy_disks)

    def cleanup(self, context, instance, network_info, block_device_info=None,
                destroy_disks=True, migrate_data=None, destroy_vifs=True):
        """Cleanup after instance being destroyed by Hypervisor."""
        container_id = self._get_container_id(instance)
        if not container_id:
            self.unplug_vifs(instance, network_info)
            return
        network.teardown_network(container_id)
        self.unplug_vifs(instance, network_info)
        self._cleanup_instance_file(instance['uuid'])

    def reboot(self, context, instance, network_info, reboot_type,
               block_device_info=None, bad_volumes_callback=None):
        container_id = self._get_container_id(instance)
        if not container_id:
            return
        self._stop(container_id, instance)
        try:
            network.teardown_network(container_id)
            if network_info:
                self.unplug_vifs(instance, network_info)
        except Exception as e:
            LOG.warning('Cannot destroy the container network'
                        ' during reboot {0}'.format(e), exc_info=True)
            return

        self.docker.start(container_id)
        try:
            if network_info:
                self.plug_vifs(instance, network_info)
                self._attach_vifs(instance, network_info)
        except Exception as e:
            LOG.warning('Cannot setup network on reboot: {0}', e,
                        exc_info=True)
            return

    def power_on(self, context, instance, network_info,
                 block_device_info=None):
        container_id = self._get_container_id(instance)
        if not container_id:
            return
        self.docker.start(container_id)
        if not network_info:
            return
        try:
            self.plug_vifs(instance, network_info)
            self._attach_vifs(instance, network_info)
        except Exception as e:
            LOG.debug('Cannot setup network: %s',
                      e, instance=instance, exc_info=True)
            msg = _('Cannot setup network: {0}')
            self.docker.kill(container_id)
            self.docker.remove_container(container_id, force=True)
            raise exception.InstanceDeployFailure(msg.format(e),
                                                  instance_id=instance['name'])

    def power_off(self, instance, timeout=0, retry_interval=0):
        container_id = self._get_container_id(instance)
        if not container_id:
            return
        self._stop(container_id, instance, timeout)

    def pause(self, instance):
        """Pause the specified instance.

        :param instance: nova.objects.instance.Instance
        """
        try:
            cont_id = self._get_container_id(instance)
            if not self.docker.pause(cont_id):
                raise exception.NovaException
        except Exception as e:
            LOG.debug('Error pause container: %s',
                      e, instance=instance, exc_info=True)
            msg = _('Cannot pause container: {0}')
            raise exception.NovaException(msg.format(e),
                                          instance_id=instance['name'])

    def unpause(self, instance):
        """Unpause paused VM instance.

        :param instance: nova.objects.instance.Instance
        """
        try:
            cont_id = self._get_container_id(instance)
            if not self.docker.unpause(cont_id):
                raise exception.NovaException
        except Exception as e:
            LOG.debug('Error unpause container: %s',
                      e, instance=instance, exc_info=True)
            msg = _('Cannot unpause container: {0}')
            raise exception.NovaException(msg.format(e),
                                          instance_id=instance['name'])

    def _get_cpu_shares(self, instance):
        """Get allocated CPUs from configured flavor.

        Docker/lxc supports relative CPU allocation.

        cgroups specifies following:
         /sys/fs/cgroup/lxc/cpu.shares = 1024
         /sys/fs/cgroup/cpu.shares = 1024

        For that reason we use 1024 as multiplier.
        This multiplier allows to divide the CPU
        resources fair with containers started by
        the user (e.g. docker registry) which has
        the default CpuShares value of zero.
        """
        if isinstance(instance, objects.Instance):
            flavor = instance.get_flavor()
        else:
            flavor = flavors.extract_flavor(instance)
        return int(flavor['vcpus']) * 1024

    def _create_container(self, instance, image_name, args):
        name = self._get_container_name(instance)
        hostname = args.pop('hostname', None)
        cpu_shares = args.pop('cpu_shares', None)
        network_disabled = args.pop('network_disabled', False)
        host_config = self.docker.create_host_config(**args)
        return self.docker.create_container(image_name,
                                            name=self._encode_utf8(name),
                                            hostname=hostname,
                                            cpu_shares=cpu_shares,
                                            network_disabled=network_disabled,
                                            host_config=host_config)

    def get_host_uptime(self):
        return hostutils.sys_uptime()



#    Copyright 2017 ARM Holdings.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_serialization import jsonutils
import pecan
import six

from zun.api.controllers import base
from zun.api.controllers import link
from zun.api.controllers.v1 import collection
from zun.api.controllers.v1.schemas import capsules as schema
from zun.api.controllers.v1.schemas import parameter_types
from zun.api.controllers.v1.views import capsules_view as view
from zun.api import utils as api_utils
from zun.api import validation
from zun.api.validation import validators
from zun.common import consts
from zun.common import exception
from zun.common.i18n import _
from zun.common import name_generator
from zun.common import policy
from zun.common import utils
import zun.conf
from zun import objects
from zun.volume import cinder_api as cinder


CONF = zun.conf.CONF
LOG = logging.getLogger(__name__)


def _get_capsule(capsule_ident):
    """Get capsule by name or UUID"""
    capsule = api_utils.get_resource('Capsule', capsule_ident)
    if not capsule:
        pecan.abort(404, ('Not found; the capsule you requested '
                          'does not exist.'))
    return capsule


def check_policy_on_capsule(capsule, action):
    context = pecan.request.context
    policy.enforce(context, action, capsule, action=action)


def check_capsule_template(tpl):
    # TODO(kevinz): add volume spec check
    tpl_json = tpl
    if isinstance(tpl, six.string_types):
        try:
            tpl_json = jsonutils.loads(tpl)
        except Exception as e:
            raise exception.FailedParseStringToJson(e)

        validator = validators.SchemaValidator(
            parameter_types.capsule_template)
        validator.validate(tpl_json)

    kind_field = tpl_json.get('kind')
    if kind_field not in ['capsule', 'Capsule']:
        raise exception.InvalidCapsuleTemplate("kind fields need to be "
                                               "set as capsule or Capsule")

    spec_field = tpl_json.get('spec')
    if spec_field is None:
        raise exception.InvalidCapsuleTemplate("No Spec found")
    # Align the Capsule restartPolicy with container restart_policy
    # Also change the template filed name from Kubernetes type to OpenStack
    # type.
    if 'restartPolicy' in spec_field.keys():
        spec_field['restartPolicy'] = \
            utils.VALID_CAPSULE_RESTART_POLICY[spec_field['restartPolicy']]
        spec_field[utils.VALID_CAPSULE_FIELD['restartPolicy']] = \
            spec_field.pop('restartPolicy')
    if spec_field.get('containers') is None:
        raise exception.InvalidCapsuleTemplate("No valid containers field")
    return spec_field, tpl_json


class CapsuleCollection(collection.Collection):
    """API representation of a collection of Capsules."""

    fields = {
        'capsules',
        'next'
    }

    """A list containing capsules objects"""

    def __init__(self, **kwargs):
        self._type = 'capsules'

    @staticmethod
    def convert_with_links(rpc_capsules, limit, url=None,
                           expand=False, legacy_api_version=False, **kwargs):
        context = pecan.request.context
        collection = CapsuleCollection()
        collection.capsules = \
            [view.format_capsule(url, p, context,
                                 legacy_api_version=legacy_api_version)
             for p in rpc_capsules]
        collection.next = collection.get_next(limit, url=url, **kwargs)
        return collection


class CapsuleController(base.Controller):
    """Controller for Capsules"""

    _custom_actions = {

    }

    @base.Controller.api_version("1.1", "1.31")
    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def get_all(self, **kwargs):
        '''Retrieve a list of capsules.'''
        return self._do_get_all(legacy_api_version=True, **kwargs)

    @base.Controller.api_version("1.32")  # noqa
    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def get_all(self, **kwargs):
        '''Retrieve a list of capsules.'''
        return self._do_get_all(**kwargs)

    def _do_get_all(self, legacy_api_version=False, **kwargs):
        context = pecan.request.context
        policy.enforce(context, "capsule:get_all",
                       action="capsule:get_all")
        if utils.is_all_projects(kwargs):
            context.all_projects = True
        limit = api_utils.validate_limit(kwargs.get('limit'))
        sort_dir = api_utils.validate_sort_dir(kwargs.get('sort_dir', 'asc'))
        sort_key = kwargs.get('sort_key', 'id')
        resource_url = kwargs.get('resource_url')
        expand = kwargs.get('expand')
        filters = None
        marker_obj = None
        marker = kwargs.get('marker')
        if marker:
            marker_obj = objects.Capsule.get_by_uuid(context,
                                                     marker)
        capsules = objects.Capsule.list(context,
                                        limit,
                                        marker_obj,
                                        sort_key,
                                        sort_dir,
                                        filters=filters)

        return CapsuleCollection.convert_with_links(
            capsules, limit, url=resource_url, expand=expand,
            sort_key=sort_key, sort_dir=sort_dir,
            legacy_api_version=legacy_api_version)

    @base.Controller.api_version("1.1", "1.31")
    @pecan.expose('json')
    @api_utils.enforce_content_types(['application/json'])
    @exception.wrap_pecan_controller_exception
    @validation.validated(schema.capsule_create)
    def post(self, **capsule_dict):
        """Create a new capsule.

        :param capsule_dict: a capsule within the request body.
        """
        return self._do_post(legacy_api_version=True, **capsule_dict)

    @base.Controller.api_version("1.32")  # noqa
    @pecan.expose('json')
    @api_utils.enforce_content_types(['application/json'])
    @exception.wrap_pecan_controller_exception
    @validation.validated(schema.capsule_create)
    def post(self, **capsule_dict):
        """Create a new capsule.

        :param capsule_dict: a capsule within the request body.
        """
        return self._do_post(**capsule_dict)

    def _do_post(self, legacy_api_version=False, **capsule_dict):
        context = pecan.request.context
        compute_api = pecan.request.compute_api
        policy.enforce(context, "capsule:create",
                       action="capsule:create")

        # Abstract the capsule specification
        capsules_template = capsule_dict.get('template')

        spec_content, template_json = \
            check_capsule_template(capsules_template)

        containers_spec, init_containers_spec = \
            utils.capsule_get_container_spec(spec_content)
        volumes_spec = utils.capsule_get_volume_spec(spec_content)

        # Create the capsule Object
        new_capsule = objects.Capsule(context, **capsule_dict)
        new_capsule.project_id = context.project_id
        new_capsule.user_id = context.user_id
        new_capsule.status = consts.CREATING
        new_capsule.create(context)
        new_capsule.volumes = []
        capsule_need_cpu = 0
        capsule_need_memory = 0
        container_volume_requests = []

        if spec_content.get('restart_policy'):
            capsule_restart_policy = spec_content.get('restart_policy')
        else:
            # NOTE(hongbin): this is deprecated but we need to maintain
            # backward-compatibility. Will remove this branch in the future.
            capsule_restart_policy = template_json.get('restart_policy',
                                                       'always')
        container_restart_policy = {"MaximumRetryCount": "0",
                                    "Name": capsule_restart_policy}
        new_capsule.restart_policy = container_restart_policy

        metadata_info = template_json.get('metadata', None)
        requested_networks_info = template_json.get('nets', [])
        requested_networks = \
            utils.build_requested_networks(context, requested_networks_info)

        if metadata_info:
            new_capsule.name = metadata_info.get('name', None)
            new_capsule.labels = metadata_info.get('labels', None)

        # create the capsule in DB so that it generates a 'id'
        new_capsule.save()

        extra_spec = {}
        az_info = template_json.get('availabilityZone')
        if az_info:
            extra_spec['availability_zone'] = az_info

        new_capsule.image = CONF.sandbox_image
        new_capsule.image_driver = CONF.sandbox_image_driver

        # calculate capsule cpu/ram
        # 1. sum all cpu/ram of regular containers
        for container_dict in containers_spec:
            if not container_dict.get('resources'):
                continue
            allocation = container_dict['resources']['requests']
            if allocation.get('cpu'):
                capsule_need_cpu += allocation['cpu']
            if allocation.get('memory'):
                capsule_need_memory += allocation['memory']
        # 2. take the maximum of each init container
        for container_dict in init_containers_spec:
            if not container_dict.get('resources'):
                continue
            allocation = container_dict['resources']['requests']
            if allocation.get('cpu'):
                capsule_need_cpu = max(capsule_need_cpu, allocation['cpu'])
            if allocation.get('memory'):
                capsule_need_memory = max(capsule_need_memory,
                                          allocation['memory'])

        merged_containers_spec = init_containers_spec + containers_spec
        for container_spec in merged_containers_spec:
            if container_spec.get('image_pull_policy'):
                policy.enforce(context, "container:create:image_pull_policy",
                               action="container:create:image_pull_policy")
            container_dict = container_spec
            container_dict['project_id'] = context.project_id
            container_dict['user_id'] = context.user_id
            name = self._generate_name_for_capsule_container(new_capsule)
            container_dict['name'] = name

            if container_dict.get('args') and container_dict.get('command'):
                container_dict['command'] = \
                    container_dict['command'] + container_dict['args']
                container_dict.pop('args')
            elif container_dict.get('args'):
                container_dict['command'] = container_dict['args']
                container_dict.pop('args')

            if container_dict.get('ports'):
                exposed_ports = {}
                ports = container_dict.pop('ports')
                for port in ports:
                    container_port = "%s/%s" % (
                        port['containerPort'],
                        port.get('protocol', 'tcp').lower())
                    host_port = {}
                    exposed_ports[container_port] = host_port
                container_dict['exposed_ports'] = exposed_ports

            if container_dict.get('resources'):
                resources_list = container_dict.get('resources')
                allocation = resources_list.get('requests')
                if allocation.get('cpu'):
                    container_dict['cpu'] = allocation.get('cpu')
                if allocation.get('memory'):
                    container_dict['memory'] = str(allocation['memory'])
                container_dict.pop('resources')

            container_dict['image_pull_policy'] = (
                container_dict.get('image_pull_policy', 'always').lower())
            container_dict['status'] = consts.CREATING
            container_dict['capsule_id'] = new_capsule.id
            container_dict['restart_policy'] = container_restart_policy
            if container_spec in init_containers_spec:
                if capsule_restart_policy == "always":
                    container_restart_policy = {"MaximumRetryCount": "10",
                                                "Name": "on-failure"}
                    container_dict['restart_policy'] = container_restart_policy
                utils.check_for_restart_policy(container_dict)
                new_container = objects.CapsuleInitContainer(context,
                                                             **container_dict)
            else:
                utils.check_for_restart_policy(container_dict)
                new_container = objects.CapsuleContainer(context,
                                                         **container_dict)
            new_container.create(context)

            if container_dict.get('volumeMounts'):
                for volume in container_dict['volumeMounts']:
                    volume['container_uuid'] = new_container.uuid
                    container_volume_requests.append(volume)

        # Deal with the volume support
        requested_volumes = \
            self._build_requested_volumes(context,
                                          volumes_spec,
                                          container_volume_requests,
                                          new_capsule)
        new_capsule.cpu = capsule_need_cpu
        new_capsule.memory = str(capsule_need_memory)
        new_capsule.save(context)

        kwargs = {}
        kwargs['extra_spec'] = extra_spec
        kwargs['requested_networks'] = requested_networks
        kwargs['requested_volumes'] = requested_volumes
        kwargs['run'] = True
        compute_api.container_create(context, new_capsule, **kwargs)
        # Set the HTTP Location Header
        pecan.response.location = link.build_url('capsules',
                                                 new_capsule.uuid)

        pecan.response.status = 202
        return view.format_capsule(pecan.request.host_url, new_capsule,
                                   context,
                                   legacy_api_version=legacy_api_version)

    @base.Controller.api_version("1.1", "1.31")
    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def get_one(self, capsule_ident):
        """Retrieve information about the given capsule.

        :param capsule_ident: UUID or name of a capsule.
        """
        return self._do_get_one(capsule_ident, legacy_api_version=True)

    @base.Controller.api_version("1.32")  # noqa
    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def get_one(self, capsule_ident):
        """Retrieve information about the given capsule.

        :param capsule_ident: UUID or name of a capsule.
        """
        return self._do_get_one(capsule_ident)

    def _do_get_one(self, capsule_ident, legacy_api_version=False):
        context = pecan.request.context
        capsule = _get_capsule(capsule_ident)
        check_policy_on_capsule(capsule.as_dict(), "capsule:get")
        return view.format_capsule(pecan.request.host_url, capsule, context,
                                   legacy_api_version=legacy_api_version)

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def delete(self, capsule_ident, **kwargs):
        """Delete a capsule.

        :param capsule_ident: UUID or Name of a capsule.
        """
        context = pecan.request.context
        if utils.is_all_projects(kwargs):
            policy.enforce(context, "capsule:delete_all_projects",
                           action="capsule:delete_all_projects")
            context.all_projects = True
        capsule = _get_capsule(capsule_ident)
        check_policy_on_capsule(capsule.as_dict(), "capsule:delete")
        compute_api = pecan.request.compute_api
        capsule.task_state = consts.CONTAINER_DELETING
        capsule.save(context)
        compute_api.container_stop(context, capsule, 10)
        compute_api.container_delete(context, capsule)
        pecan.response.status = 204

    def _generate_name_for_capsule_container(self, new_capsule):
        """Generate a random name like: zeta-22-container."""
        name_gen = name_generator.NameGenerator()
        name = name_gen.generate()
        if new_capsule.name is None:
            return 'capsule-' + new_capsule.uuid + '-' + name
        else:
            return 'capsule-' + new_capsule.name + '-' + name

    def _build_requested_volumes(self, context, volume_spec,
                                 volume_mounts, capsule):
        # NOTE(hongbin): We assume cinder is the only volume provider here.
        # The logic needs to be re-visited if a second volume provider
        # (i.e. Manila) is introduced.
        # NOTE(kevinz): We assume the volume_mounts has been pretreated,
        # there won't occur that volume multiple attach and no untapped
        # volume.
        cinder_api = cinder.CinderAPI(context)
        volume_driver = "cinder"
        requested_volumes = {}
        volume_created = []
        try:
            for mount in volume_spec:
                mount_driver = mount[volume_driver]
                auto_remove = False
                if mount_driver.get("volumeID"):
                    uuid = mount_driver.get("volumeID")
                    volume = cinder_api.search_volume(uuid)
                    cinder_api.ensure_volume_usable(volume)
                else:
                    size = mount_driver.get("size")
                    volume = cinder_api.create_volume(size)
                    volume_created.append(volume)
                    if "autoRemove" in mount_driver.keys() \
                            and mount_driver.get("autoRemove", False):
                        auto_remove = True

                mount_destination = None
                container_uuid = None

                volume_object = objects.Volume(
                    context,
                    cinder_volume_id=volume.id,
                    volume_provider=volume_driver,
                    user_id=context.user_id,
                    project_id=context.project_id,
                    auto_remove=auto_remove)
                volume_object.create(context)

                for item in volume_mounts:
                    if item['name'] == mount['name']:
                        mount_destination = item['mountPath']
                        container_uuid = item['container_uuid']
                        volmapp = objects.VolumeMapping(
                            context,
                            container_path=mount_destination,
                            user_id=context.user_id,
                            project_id=context.project_id,
                            volume_id=volume_object.id)
                        requested_volumes.setdefault(container_uuid, [])
                        requested_volumes[container_uuid].append(volmapp)

                if not mount_destination or not container_uuid:
                    msg = _("volume mount parameters is invalid.")
                    raise exception.Invalid(msg)
        except Exception as e:
            # if volume search or created failed, will remove all
            # the created volume. The existed volume will remain.
            for volume in volume_created:
                try:
                    cinder_api.delete_volume(volume.id)
                except Exception as exc:
                    LOG.error('Error on deleting volume "%s": %s.',
                              volume.id, six.text_type(exc))

            # Since the container and capsule database model has been created,
            # we need to delete them here due to the volume create failed.
            for container in capsule.containers:
                try:
                    container.destroy(context)
                except Exception as exc:
                    LOG.warning('fail to delete the container %s: %s',
                                container.uuid, exc)

            capsule.destroy(context)

            raise e

        return requested_volumes



# Copyright 2017 OpenStack Foundation  # All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Fake Loadable subclasses module #1
"""

from zun.tests.unit.scheduler import fake_loadables


class FakeLoadableSubClass1(fake_loadables.FakeLoadable):
    pass


class FakeLoadableSubClass2(fake_loadables.FakeLoadable):
    pass


class _FakeLoadableSubClass3(fake_loadables.FakeLoadable):
    """Classes beginning with '_' will be ignored."""
    pass


class FakeLoadableSubClass4(object):
    """Not a correct subclass."""


def return_valid_classes():
    return [FakeLoadableSubClass1, FakeLoadableSubClass2]


def return_invalid_classes():
    return [FakeLoadableSubClass1, _FakeLoadableSubClass3,
            FakeLoadableSubClass4]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests for base filter
"""
import inspect

import mock

from zun.scheduler import base_filters
from zun.tests import base


class BaseFilterTestCase(base.TestCase):
    """Test case for base filter class."""

    @mock.patch('zun.scheduler.base_filters.BaseFilter._filter_one')
    def test_filter_all(self, mock_filter_one):
        mock_filter_one.side_effect = [True, False, True]
        filter_obj_list = ['obj1', 'obj2', 'obj3']
        container = {}
        base_filter = base_filters.BaseFilter()
        extra_spec = {}
        result = base_filter.filter_all(filter_obj_list, container, extra_spec)
        self.assertTrue(inspect.isgenerator(result))
        self.assertEqual(['obj1', 'obj3'], list(result))

    def test_run_filter_for_index(self):
        base_filter = base_filters.BaseFilter()
        base_filter.run_filter_once_per_request = False
        result = base_filter.run_filter_for_index(2)
        self.assertTrue(result)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import copy

from oslo_log import log as logging
import retrying

from zun.common import consts
from zun.common import exception
from zun.common import utils
from zun.compute import claims
import zun.conf
from zun import objects
from zun.objects import base as obj_base
from zun.pci import manager as pci_manager
from zun.scheduler.client import query as scheduler_client


CONF = zun.conf.CONF
LOG = logging.getLogger(__name__)
COMPUTE_RESOURCE_SEMAPHORE = "compute_resources"


class ComputeNodeTracker(object):
    def __init__(self, host, container_driver, reportclient):
        self.host = host
        self.container_driver = container_driver
        self.compute_node = None
        self.tracked_containers = {}
        self.old_resources = collections.defaultdict(objects.ComputeNode)
        self.scheduler_client = scheduler_client.SchedulerClient()
        self.pci_tracker = None
        self.reportclient = reportclient
        self.rp_uuid = None

    def _setup_pci_tracker(self, context, compute_node):
        if not self.pci_tracker:
            n_id = compute_node.uuid
            self.pci_tracker = pci_manager.PciDevTracker(context, node_id=n_id)
            dev_json = self.container_driver.get_pci_resources()
            self.pci_tracker.update_devices_from_compute_resources(dev_json)

            dev_pools_obj = self.pci_tracker.stats.to_device_pools_obj()
            compute_node.pci_device_pools = dev_pools_obj

    def update_available_resources(self, context):
        resources = self.container_driver.get_available_resources()
        # We allow 'cpu_used' to be missing from the container driver,
        # but the DB requires it to be non-null so just initialize it to 0.
        resources.setdefault('cpu_used', 0)

        # Check if the compute_node is already registered
        node = self._get_compute_node(context)
        if not node:
            # If not, register it and pass the object to the driver
            node = objects.ComputeNode(context)
            node.hostname = self.host
            self._copy_resources(node, resources)
            node.create(context)
            LOG.info('Node created for :%(host)s', {'host': self.host})
        else:
            self._copy_resources(node, resources)
        node.rp_uuid = self._get_node_rp_uuid(context, node)
        self._setup_pci_tracker(context, node)
        self.compute_node = node
        self._update_available_resource(context)
        # NOTE(sbiswas7): Consider removing the return statement if not needed
        return node

    def _copy_resources(self, node, resources):
        keys = ["numa_topology", "mem_total", "mem_free", "mem_available",
                "mem_used", "total_containers", "running_containers",
                "paused_containers", "stopped_containers", "cpus",
                "architecture", "os_type", "os", "kernel_version", "cpu_used",
                "labels", "disk_total", "disk_quota_supported", "runtimes",
                "enable_cpu_pinning"]
        for key in keys:
            if key in resources:
                setattr(node, key, resources[key])

    def _get_compute_node(self, context):
        """Returns compute node for the host"""
        try:
            return objects.ComputeNode.get_by_name(context, self.host)
        except exception.ComputeNodeNotFound:
            LOG.warning("No compute node record for: %(host)s",
                        {'host': self.host})

    def _get_node_rp_uuid(self, context, node):
        if self.rp_uuid:
            return self.rp_uuid

        if CONF.compute.host_shared_with_nova:
            try:
                self.rp_uuid = self.reportclient.get_provider_by_name(
                    context, node.hostname)['uuid']
            except exception.ResourceProviderNotFound:
                raise exception.ComputeHostNotFound(host=node.nodename)
        else:
            self.rp_uuid = node.uuid

        return self.rp_uuid

    @utils.synchronized(COMPUTE_RESOURCE_SEMAPHORE)
    def container_claim(self, context, container, pci_requests, limits=None):
        """Indicate resources are needed for an upcoming container build.

        This should be called before the compute node is about to perform
        an container build operation that will consume additional resources.

        :param context: security context
        :param container: container to reserve resources for.
        :type container: zun.objects.container.Container object
        :param pci_requests: pci requests for sriov port.
        :param limits: Dict of oversubscription limits for memory, disk,
                       and CPUs.
        :returns: A Claim ticket representing the reserved resources.  It can
                  be used to revert the resource usage if an error occurs
                  during the container build.
        """
        # No memory, cpu, disk or pci_request specified, no need to claim
        # resource now.
        if not (container.memory or container.cpu or pci_requests or
                container.disk):
            return claims.NopClaim()

        # We should have the compute node created here, just get it.
        self.compute_node = self._get_compute_node(context)

        claim = claims.Claim(context, container, self, self.compute_node,
                             pci_requests, limits=limits)

        if self.pci_tracker:
            self.pci_tracker.claim_container(context, container.uuid,
                                             pci_requests)

        self._set_container_host(context, container)
        self._update_usage_from_container(context, container)
        # persist changes to the compute node:
        self._update(context, self.compute_node)

        return claim

    @utils.synchronized(COMPUTE_RESOURCE_SEMAPHORE)
    def container_update_claim(self, context, new_container, old_container,
                               limits=None):
        """Indicate resources are needed for an upcoming container update.

        This should be called before the compute node is about to perform
        an container update operation that will consume additional resources.

        :param context: security context
        :param new_container: container to be updated to.
        :type new_container: zun.objects.container.Container object
        :param old_container: container to be updated from.
        :type old_container: zun.objects.container.Container object
        :param limits: Dict of oversubscription limits for memory, disk,
                       and CPUs.
        :returns: A Claim ticket representing the reserved resources.  It can
                  be used to revert the resource usage if an error occurs
                  during the container update.
        """
        if (new_container.cpu == old_container.cpu and
                new_container.memory == old_container.memory):
            return claims.NopClaim()

        # We should have the compute node created here, just get it.
        self.compute_node = self._get_compute_node(context)

        claim = claims.UpdateClaim(context, new_container, old_container,
                                   self, self.compute_node, limits=limits)

        self._update_usage_from_container_update(context, new_container,
                                                 old_container)
        # persist changes to the compute node:
        self._update(context, self.compute_node)

        return claim

    def disabled(self, hostname):
        if not self.compute_node:
            return True

        return (hostname != self.compute_node.hostname or
                not self.container_driver.node_is_available(hostname))

    def _set_container_host(self, context, container):
        """Tag the container as belonging to this host.

        This should be done while the COMPUTE_RESOURCES_SEMAPHORE is held so
        the resource claim will not be lost if the audit process starts.
        """
        container.host = self.host
        container.save(context)

    def _update_usage_from_container(self, context, container,
                                     is_removed=False):
        """Update usage for a single container."""

        uuid = container.uuid
        is_new_container = uuid not in self.tracked_containers
        is_removed_container = not is_new_container and is_removed

        if is_new_container:
            self.tracked_containers[uuid] = \
                obj_base.obj_to_primitive(container)
            sign = 1

        if is_removed_container:
            self.tracked_containers.pop(uuid)
            sign = -1

        if is_new_container or is_removed_container:
            if self.pci_tracker:
                self.pci_tracker.update_pci_for_container(context, container,
                                                          sign=sign)

            # new container, update compute node resource usage:
            self._update_usage(self._get_usage_dict(container), sign=sign)

    def _update_usage_from_container_update(self, context, new_container,
                                            old_container):
        """Update usage for a container update."""
        uuid = new_container.uuid
        self.tracked_containers[uuid] = obj_base.obj_to_primitive(
            new_container)
        # update compute node resource usage
        self._update_usage(self._get_usage_dict(old_container), sign=-1)
        self._update_usage(self._get_usage_dict(new_container))

    def _update_usage_from_containers(self, context, containers):
        """Calculate resource usage based on container utilization.

        This is different than the container daemon view as it will account
        for all containers assigned to the local compute host, even if they
        are not currently powered on.
        """
        self.tracked_containers.clear()

        cn = self.compute_node
        # set some initial values, reserve room for host
        cn.cpu_used = 0
        cn.mem_free = cn.mem_total
        cn.mem_used = 0
        cn.running_containers = 0
        cn.disk_used = 0

        for cnt in containers:
            self._update_usage_from_container(context, cnt)

        cn.mem_free = max(0, cn.mem_free)

    def _update_usage(self, usage, sign=1):
        mem_usage = usage['memory']
        cpus_usage = usage.get('cpu', 0)
        disk_usage = usage['disk']
        cpuset_cpus_usage = None
        numa_node_id = 0
        if 'cpuset_cpus' in usage.keys():
            cpuset_cpus_usage = usage['cpuset_cpus']
            numa_node_id = usage['node']

        cn = self.compute_node
        numa_topology = cn.numa_topology.nodes
        cn.mem_used += sign * mem_usage
        cn.cpu_used += sign * cpus_usage
        cn.disk_used += sign * disk_usage

        # free ram may be negative, depending on policy:
        cn.mem_free = cn.mem_total - cn.mem_used

        cn.running_containers += sign * 1

        if cpuset_cpus_usage:
            for numa_node in numa_topology:
                if numa_node.id == numa_node_id:
                    numa_node.mem_available = (numa_node.mem_available -
                                               mem_usage * sign)
                    if sign > 0:
                        numa_node.pin_cpus(cpuset_cpus_usage)
                        cn._changed_fields.add('numa_topology')
                    else:
                        numa_node.unpin_cpus(cpuset_cpus_usage)
                        cn._changed_fields.add('numa_topology')

    def _update(self, context, compute_node):
        if not self._resource_change(compute_node):
            return
        # Persist the stats to the Scheduler
        self.scheduler_client.update_resource(compute_node)

        self._update_to_placement(context, compute_node)

        if self.pci_tracker:
            self.pci_tracker.save()

    def _resource_change(self, compute_node):
        """Check to see if any resources have changed."""
        hostname = compute_node.hostname
        old_compute = self.old_resources[hostname]
        if not obj_base.obj_equal_prims(
                compute_node, old_compute, ['updated_at']):
            self.old_resources[hostname] = copy.deepcopy(compute_node)
            return True
        return False

    @retrying.retry(stop_max_attempt_number=4,
                    retry_on_exception=lambda e: isinstance(
                        e, exception.ResourceProviderUpdateConflict))
    def _update_to_placement(self, context, compute_node):
        """Send resource and inventory changes to placement."""
        nodename = compute_node.hostname
        node_rp_uuid = self._get_node_rp_uuid(context, compute_node)
        # Persist the stats to the Scheduler
        # First try update_provider_tree
        # Retrieve the provider tree associated with this compute node.  If
        # it doesn't exist yet, this will create it with a (single, root)
        # provider corresponding to the compute node.
        prov_tree = self.reportclient.get_provider_tree_and_ensure_root(
            context, node_rp_uuid, name=compute_node.hostname)
        # Let the container driver rearrange the provider tree and set/update
        # the inventory, traits, and aggregates throughout.
        self.container_driver.update_provider_tree(prov_tree, nodename)
        # Inject driver capabilities traits into the provider
        # tree.  We need to determine the traits that the container
        # driver owns - so those that come from the tree itself
        # (via the container driver) plus the compute capabilities
        # traits, and then merge those with the traits set
        # externally that the driver does not own - and remove any
        # set on the provider externally that the driver owns but
        # aren't in the current list of supported traits.  For
        # example, let's say we reported multiattach support as a
        # trait at t1 and then at t2 it's not, so we need to
        # remove it.  But at both t1 and t2 there is a
        # CUSTOM_VENDOR_TRAIT_X which we can't touch because it
        # was set externally on the provider.
        # We also want to sync the COMPUTE_STATUS_DISABLED trait based
        # on the related zun-compute service's disabled status.
        traits = self._get_traits(
            context, nodename, provider_tree=prov_tree)
        prov_tree.update_traits(nodename, traits)

        self.reportclient.update_from_provider_tree(context, prov_tree)

    def _get_traits(self, context, nodename, provider_tree):
        """Synchronizes internal and external traits for the node provider.

        This works in conjunction with the ComptueDriver.update_provider_tree
        flow and is used to synchronize traits reported by the compute driver,
        traits based on information in the ComputeNode record, and traits set
        externally using the placement REST API.

        :param context: RequestContext for cell database access
        :param nodename: ComputeNode.hostname for the compute node
            resource provider whose traits are being synchronized; the node
            must be in the ProviderTree.
        :param provider_tree: ProviderTree being updated
        """
        # Get the traits from the ProviderTree which will be the set
        # of driver-owned traits plus any externally defined traits set
        # on the provider that aren't owned by the container driver.
        traits = provider_tree.data(nodename).traits

        # Now get the driver's capabilities and add any supported
        # traits that are missing, and remove any existing set traits
        # that are not currently supported.
        capabilities_traits = self.container_driver.capabilities_as_traits()
        for trait, supported in capabilities_traits.items():
            if supported:
                traits.add(trait)
            elif trait in traits:
                traits.remove(trait)

        self._sync_compute_service_disabled_trait(context, traits)

        return list(traits)

    def _sync_compute_service_disabled_trait(self, context, traits):
        """Synchronize the ZUN_COMPUTE_STATUS_DISABLED  trait on the node provider.

        Determines if the ZUN_COMPUTE_STATUS_DISABLED trait should be added to
        or removed from the provider's set of traits based on the related
        zun-compute service disabled status.

        :param context: RequestContext for cell database access
        :param traits: set of traits for the compute node resource provider;
            this is modified by reference
        """
        trait = consts.ZUN_COMPUTE_STATUS_DISABLED
        try:
            service = objects.ZunService.get_by_host_and_binary(
                context, self.host, 'zun-compute')
            if service.disabled:
                # The service is disabled so make sure the trait is reported.
                traits.add(trait)
            else:
                # The service is not disabled so do not report the trait.
                traits.discard(trait)
        except exception.NotFound:
            # This should not happen but handle it gracefully. The scheduler
            # should ignore this node if the compute service record is gone.
            LOG.error('Unable to find services table record for zun-compute '
                      'host %s', self.host)

    @utils.synchronized(COMPUTE_RESOURCE_SEMAPHORE)
    def _update_available_resource(self, context):

        # if we could not init the compute node the tracker will be
        # disabled and we should quit now
        if self.disabled(self.host):
            return

        # Grab all containers assigned to this node:
        containers = objects.Container.list_by_host(context, self.host)

        # Now calculate usage based on container utilization:
        self._update_usage_from_containers(context, containers)

        # No migration for docker, is there will be orphan container? Nova has.

        cn = self.compute_node

        # update the compute_node
        self._update(context, cn)
        LOG.debug('Compute_service record updated for %(host)s',
                  {'host': self.host})

    def _get_usage_dict(self, container, **updates):
        """Make a usage dict _update methods expect.

        Accepts an Container, and a set of updates.
        Converts the object to a dict and applies the updates.

        :param container: container as an object
        :param updates: key-value pairs to update the passed object.

        :returns: a dict with all the information from container updated
                  with updates
        """
        usage = {}
        # (Fixme): The Container.memory is string.
        memory = 0
        if container.memory:
            memory = int(container.memory)
        usage = {'memory': memory,
                 'cpu': container.cpu or 0,
                 'disk': container.disk or 0}
        if container.cpuset.cpuset_cpus:
            usage['cpuset_cpus'] = container.cpuset.cpuset_cpus
            usage['node'] = int(container.cpuset.cpuset_mems)

        return usage

    @utils.synchronized(COMPUTE_RESOURCE_SEMAPHORE)
    def abort_container_claim(self, context, container):
        """Remove usage from the given container."""
        self._update_usage_from_container(context, container, is_removed=True)

        self._update(context, self.compute_node)

    @utils.synchronized(COMPUTE_RESOURCE_SEMAPHORE)
    def abort_container_update_claim(self, context, new_container,
                                     old_container):
        """Remove usage from the given container."""
        self._update_usage_from_container_update(context, old_container,
                                                 new_container)
        self._update(self.compute_node)

    @utils.synchronized(COMPUTE_RESOURCE_SEMAPHORE)
    def remove_usage_from_container(self, context, container,
                                    is_removed=True):
        """Just a wrapper of the private function to hold lock."""

        # We need to get the latest compute node info
        self.compute_node = self._get_compute_node(context)
        self._update_usage_from_container(context, container, is_removed)
        self._update(context, self.compute_node)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Chance (Random) Scheduler implementation
"""

import random

from zun.common import exception
from zun.common.i18n import _
from zun.scheduler import driver


class ChanceScheduler(driver.Scheduler):
    """Implements Scheduler as a random node selector."""

    def _schedule(self, context):
        """Picks a host that is up at random."""
        hosts = self.hosts_up(context)
        if not hosts:
            msg = _("Is the appropriate service running?")
            raise exception.NoValidHost(reason=msg)

        return random.choice(hosts)

    def select_destinations(self, context, containers, extra_spec):
        """Selects random destinations."""
        dests = []
        for container in containers:
            host = self._schedule(context)
            host_state = dict(host=host, nodename=None, limits=None)
            dests.append(host_state)

        if len(dests) < 1:
            reason = _('There are not enough hosts available.')
            raise exception.NoValidHost(reason=reason)

        return dests



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from testtools.matchers import HasLength

from zun import objects
from zun.tests.unit.db import base
from zun.tests.unit.db import utils


class TestNetworkObject(base.DbTestCase):

    def setUp(self):
        super(TestNetworkObject, self).setUp()
        self.fake_network = utils.get_test_network()

    def test_create(self):
        with mock.patch.object(self.dbapi, 'create_network',
                               autospec=True) as mock_create_network:
            mock_create_network.return_value = self.fake_network
            network = objects.Network(self.context, **self.fake_network)
            network.create(self.context)
            mock_create_network.assert_called_once_with(self.context,
                                                        self.fake_network)
            self.assertEqual(self.context, network._context)

    def test_save(self):
        uuid = self.fake_network['uuid']
        with mock.patch.object(self.dbapi, 'get_network_by_uuid',
                               autospec=True) as mock_get_network:
            mock_get_network.return_value = self.fake_network
            with mock.patch.object(self.dbapi, 'update_network',
                                   autospec=True) as mock_update_network:
                network = objects.Network.get_by_uuid(self.context, uuid)
                network.name = 'network-test'
                network.neutron_net_id = 'test-id'
                network.save()
                mock_get_network.assert_called_once_with(self.context, uuid)
                params = {'name': 'network-test', 'neutron_net_id': 'test-id'}
                mock_update_network.assert_called_once_with(None,
                                                            uuid,
                                                            params)
                self.assertEqual(self.context, network._context)

    def test_list(self):
        with mock.patch.object(self.dbapi, 'list_networks',
                               autospec=True) as mock_get_list:
            mock_get_list.return_value = [self.fake_network]
            networks = objects.Network.list(self.context)
            self.assertEqual(1, mock_get_list.call_count)
            self.assertThat(networks, HasLength(1))
            self.assertIsInstance(networks[0], objects.Network)
            self.assertEqual(self.context, networks[0]._context)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

quota_group = cfg.OptGroup(
    name='quota',
    title='Quota Options',
    help="""
Quota options allow to manage quotas in openstack zun deployment.
""")

quota_opts = [
    cfg.IntOpt(
        'containers',
        min=-1,
        default=40,
        help="""
The number of containers allowed per project.

Possible values

* A positive integer or 0.
* -1 to disable the quota.
"""),
    cfg.IntOpt(
        'memory',
        min=-1,
        default=50 * 1024,
        help="""
The number of megabytes of container RAM allowed per project.

Possible values:

* A positive integer or 0.
* -1 to disable the quota.
"""
    ),
    cfg.IntOpt(
        'cpu',
        min=-1,
        default=20,
        help="""
The number of container cores or vCPUs allowed per project.

Possitive values:

* A positive integer or 0.
* -1 to disable the quota.
"""
    ),
    cfg.IntOpt(
        'disk',
        min=-1,
        default=100,
        help="""
The number of gigabytes of container Disk allowed per project.

Possitive values:

* A possitive integer or 0.
* -1 to disable the quota.
"""
    ),
    cfg.StrOpt('driver',
               default='zun.common.quota.DbQuotaDriver',
               choices=('zun.common.quota.DbQuotaDriver',
                        'zun.common.quota.NoopQuotaDriver'),
               help="""
Provides abstraction for quota checks. Users can configure a specific
driver to use for quota checks.

Possible values:

* zun.common.quota.DbQuotaDriver: Stores quota limit information
  in the database and relies on te quota_* configuration options for default
  quota limit values. Counts quota usage on-demand.
* zun.common.quota.NoopQuotaDriver: Ignores quota and treats all resources as
  unlimited.
""")
]


def register_opts(conf):
    conf.register_group(quota_group)
    conf.register_opts(quota_opts, group=quota_group)


def list_opts():
    return {quota_group: quota_opts}
